from pygears import gear, alternative, module
from pygears.typing import Queue, Bool
from pygears.typing.flatten import flatten as type_flatten
from operator import and_

from functools import reduce


def flatten_func(d, lvl):
    dout_lvl = d.lvl - lvl

    if dout_lvl == 0:
        return d.data
    else:
        eot = d.eot[lvl+1:] @ Bool(reduce(and_, d.eot[:lvl+1]))
        return Queue[type(d.data), dout_lvl](d.data, eot)


@gear
async def flatten(din: Queue,
                  *,
                  lvl=1,
                  dout_lvl=b'din.lvl - lvl') -> b'Queue[din.data, dout_lvl]':
    async with din as d:
        yield flatten_func(d, lvl)


@alternative(flatten)
@gear(enablement=b'issubclass(din, Tuple)')
def flatten_tuple(din, *, lvl=1):
    return din >> type_flatten(din.dtype, lvl)



from pygears.conf import PluginBase
from .generate import vgen_generate
from .inst import vgen_inst
from .vmod import VModuleInst
from pygears.conf import PluginBase, reg

from .vcompile import compile_gear, compile_gear_body


class VGenPlugin(PluginBase):
    @classmethod
    def bind(cls):
        pass
        # reg['vgen/flow'] = [vgen_inst, vgen_generate]
        # reg['vgen/resolvers'] = 
        #     HDLFileResolver, HDLTemplateResolver, HLSResolver,
        #     HierarchicalResolver
        # ])

        # reg['svgen/dflt_resolver'] = BlackBoxResolver
        # # reg['vgen/flow'] = [vgen_inst]
        # reg['vgen/module_namespace'] = {}
        # reg['vgen/module_namespace/Gear'] = VModuleInst
        # reg['vgen/module_namespace/GearHierRoot'] = VModuleInst

__all__ = ['compile_gear', 'compile_gear_body']



from .visitor import (visit_ast, visit_block, node_visitor, Context,
                      HLSSyntaxError, Submodule, Function, FuncContext,
                      GearContext, form_hls_syntax_error)
from .. import ir
from .. import ir_utils

from . import modules
from . import async_stmts
from . import expr
from . import ctrl
from . import stmt
from . import call

__all__ = ['visit_ast', 'nodes', 'visit_block']



from pygears import gear, Intf
from pygears.lib import czip
from pygears.typing import Tuple, Uint, Union, Queue
from pygears.lib import fmap, demux, decouple, fifo, union_collapse
from pygears.lib import priority_mux, replicate

TCfg = Tuple[{'reduce_size': Uint['w_reduce_size'], 'init': 't_acc'}]


@gear
def reduce2(din, cfg: TCfg, *, f, max_size):
    """Similar to the Python reduce function, applies a rolling computation to
    sequential pairs of values in a list. The ``din`` input is of type
    :class:`Queue` which holds the values to be used for computation while the
    ``cfg`` input is a :class:`Tuple` consisting of a ``reduce_size`` field and
    the ``init`` field holding the inital value.

    Args:
        f: Function to be performed
        max_size: Maximal length of the input `Queue` which is the depth of the
          FIFO used for storing intermediate values

    Returns:
        The result of the reduce operation
    """

    acctype = cfg.dtype['init']

    qtype = Queue[acctype, din.dtype.lvl - 1]

    temp_res = Intf(dtype=qtype)
    cfg_rep = cfg | replicate
    sec_opnd = (cfg_rep, temp_res) \
        | priority_mux \
        | fmap(f=union_collapse, fcat=czip, lvl=1)

    result = czip(din, sec_opnd) | decouple | fmap(f=f, fcat=czip, lvl=2)
    acc, fin_res = result | Union[qtype, qtype] | demux
    acc | fifo(intfs=[temp_res], depth=max_size)

    return fin_res



from .base import TypingMeta


class UnitMeta(TypingMeta):
    def __int__(self):
        return 0

    @property
    def width(self):
        return 0

    def __len__(self):
        return 0

    def __str__(self):
        return '()'

    def __matmul__(self, other):
        return other

    def __rmatmul__(self, other):
        return other

    @property
    def _base(self):
        return self



class Unit(metaclass=UnitMeta):
    def __init__(self, v=None):
        if (v is not None) and (not isinstance(v, Unit)):
            raise TypeError

    def __getitem__(self, key):
        raise IndexError

    def __str__(self):
        return '()'

    def __repr__(self):
        return 'Unit()'

    def code(self):
        return 0

    def __int__(self):
        return 0

    def __len__(self):
        return 0

    def __matmul__(self, other):
        return other

    def __rmatmul__(self, other):
        return other

    def __eq__(self, other):
        return isinstance(other, Unit)

    @classmethod
    def decode(cls, val):
        return cls()

    def __hash__(self):
        return hash(type(self))



from pygears.hls import ir
from functools import partial
from pygears.typing import Array, Integer, Queue, code, typeof, Integral, Tuple, Union, Int, Unit, Uint
from .sv_keywords import sv_keywords

# TODO: Use precedence of operators to induce parenthesis around expressions

SLICE_FUNC_TEMPLATE = """function {4} [{2}:0] slice_{0}_{1}_{3}(input [{3}:0] val);
    slice_{0}_{1}_{3} = val[{0}:{1}];
endfunction
"""

INDEX_FUNC_TEMPLATE = """function {0} [{1}-1:0] index_{1}_{2}_{3}(input [{2}:0] val, input [{3}:0] ind);
    index_{1}_{2}_{3} = val[ind*{1}+:{1}];
endfunction
"""


def get_slice_func(aux_funcs, start, stop, din_width, signed):
    name = f'slice_{stop}_{start}_{din_width-1}'
    if name not in aux_funcs:
        aux_funcs[name] = SLICE_FUNC_TEMPLATE.format(stop, start, stop - start, din_width - 1,
                                                     'signed' if signed else '')

    return name


def get_index_func(aux_funcs, din_width, ind_width, part_width, signed):
    name = f'index_{part_width}_{din_width-1}_{ind_width-1}'
    if name not in aux_funcs:
        aux_funcs[name] = INDEX_FUNC_TEMPLATE.format('signed' if signed else '', part_width,
                                                     din_width - 1, ind_width - 1)

    return name


def index_to_sv_slice(dtype, key):
    subtype = dtype[key]

    if isinstance(key, slice):
        key = min(key.start, key.stop)

    if key is None or key == 0:
        low_pos = 0
    else:
        low_pos = dtype[:key].width

    high_pos = low_pos + subtype.width - 1

    return f'{high_pos}:{low_pos}'


def sieve_slices(dtype, keys):
    if not isinstance(keys, tuple):
        keys = (keys, )

    return list(
        map(partial(index_to_sv_slice, dtype),
            filter(lambda i: getattr(dtype[i], 'width', 0) > 0, keys)))


class SVExpressionVisitor:
    def __init__(self, aux_funcs=None):
        self.separator = '.'
        self.expr = svexpr

        # TODO: Think of a better way of communicating parent context
        # Used for SubscriptExpr which can have 'store' context, while its value can be 'load'
        self.ctx = None

        if aux_funcs is None:
            aux_funcs = {}

        self.aux_funcs = aux_funcs

    def visit(self, node):
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_OperandVal(self, node):
        if node.context:
            return f'{node.op.name}_{node.context}'

    def visit_ResExpr(self, node):
        if isinstance(node.val, tuple):
            res = []
            for op in reversed(node.val):
                res_op = ir.ResExpr(op)
                if res_op != ir.ResExpr(Unit()) and res_op.dtype != Uint[0]:
                    svexpr = self.visit(res_op)
                    res.append(self.cast_svexpr(svexpr, res_op.dtype, type(op)))

            if not res:
                return None

            return '{' + ', '.join(res) + '}'

        if getattr(node.val, 'unknown', False):
            return f"{node.dtype.width}'bx"

        val = node.val
        if isinstance(node.val, ir.EmptyType):
            if node.dtype is None:
                return f"'x"
            else:
                return f"{node.dtype.width}'bx"

        elif not isinstance(node.val, Integer):
            val = Integer(code(node.val, int))

        sign = '-' if val < 0 else ''
        return f"{sign}{val.width}'d{abs(int(val))}"

    def visit_FunctionCall(self, node):
        return (f'{node.name}(' + ', '.join(str(self.visit(op)) for op in node.operands) + ')')

    def visit_Interface(self, node):
        return node.name

    def visit_Variable(self, node):
        breakpoint()
        return f'{node.name}_v'

    def visit_Register(self, node):
        return f'{node.name}_v'

    def visit_Name(self, node):
        name = node.name
        if name in sv_keywords:
            name = f'pg_{name}'

        if isinstance(node.obj, ir.Variable) and node.obj.reg:
            if self.ctx == 'store' or node.ctx == 'store':
                return f'{name}_next'

        return name

    def visit_Await(self, node):
        return self.visit(node.expr)

    def visit_Component(self, node):
        if (node.field == 'data'):
            return f'{node.val.name}_s'
        else:
            return self.separator.join([self.visit(node.val), node.field])

    def visit_CallExpr(self, node):
        breakpoint()

    def visit_AttrExpr(self, node):
        val = [self.visit(node.val)]
        # if node.attr:
        #     if typeof(node.val.dtype, Queue):
        #         try:
        #             node.val.dtype[node.attr[0]]
        #         except KeyError:
        #             val.append('data')
        return self.separator.join(val + [node.attr])

    def cast_sign(self, expr, expr_dtype, cast_dtype):
        res_signed = getattr(expr_dtype, 'signed', False)
        op_signed = getattr(cast_dtype, 'signed', False)

        if res_signed != op_signed:
            sign = 'signed' if res_signed else 'unsigned'
            expr = f"{sign}'({expr})"

        return expr

    def cast_width(self, expr, expr_dtype, cast_dtype):
        if len(cast_dtype) != len(expr_dtype):
            expr = f"{expr_dtype.width}'({expr})"

        return expr

    def cast_svexpr(self, svexpr, expr_dtype, cast_dtype):
        expr_signed = getattr(expr_dtype, 'signed', False)
        res_signed = getattr(cast_dtype, 'signed', False)

        expr_width = expr_dtype.width
        cast_width = cast_dtype.width

        if cast_width == 0:
            return None

        if res_signed != expr_signed:
            if res_signed:
                svexpr = f"signed'({{1'b0, {svexpr}}})"
                expr_width += 1
            else:
                svexpr = f"unsigned'({svexpr})"

        if cast_width != expr_width:
            svexpr = f"{cast_width}'({svexpr})"

        return svexpr

    def cast_expr(self, expr, cast_dtype):
        return self.cast_svexpr(self.visit(expr), expr.dtype, cast_dtype)

    def visit_CastExpr(self, node):
        return self.cast_expr(node.operand, node.cast_to)

    def visit_ConcatExpr(self, node):
        svexprs = []
        for op in reversed(node.operands):
            if op.dtype.width == 0:
                continue

            sv = self.visit(op)
            if sv is None:
                continue
            svexprs.append(str(sv))

        if svexprs:
            return '{' + ', '.join(svexprs) + '}'

        return None

    def visit_ArrayOpExpr(self, node):
        val = self.visit(node.array)
        return f'{ir.OPMAP[node.operator]}({val})'

    def visit_UnaryOpExpr(self, node):
        val = self.visit(node.operand)

        if val is None:
            return None

        res = f'{ir.OPMAP[node.operator]}({val})'

        if node.operator in [ir.opc.Invert]:
            res = self.cast_svexpr(res, node.dtype, node.operand.dtype)
            # f"{node.dtype.width}'({res})"

        return res

    def visit_BinOpExpr(self, node):
        ops = [self.visit(op) for op in node.operands]
        op_dtypes = [op.dtype for op in node.operands]
        op_sign = [getattr(dtype, 'signed', False) for dtype in op_dtypes]

        if node.operator in [
                ir.opc.Add, ir.opc.Sub, ir.opc.Mult, ir.opc.BitOr, ir.opc.BitAnd, ir.opc.BitXor
        ]:
            cast_dtype = node.dtype
            ops = [self.cast_svexpr(expr, dtype, cast_dtype) for expr, dtype in zip(ops, op_dtypes)]
        elif node.operator == ir.opc.LShift:
            if op_dtypes[0].width < node.dtype.width:
                ops[0] = self.cast_svexpr(ops[0], op_dtypes[0], node.dtype)
        elif node.operator in [
                ir.opc.Eq, ir.opc.Gt, ir.opc.GtE, ir.opc.Lt, ir.opc.LtE, ir.opc.NotEq, ir.opc.And,
                ir.opc.Or
        ]:
            if op_sign[0] and not op_sign[1]:
                ops[1] = self.cast_svexpr(ops[1], op_dtypes[1], op_dtypes[0])
            elif op_sign[1] and not op_sign[0]:
                ops[0] = self.cast_svexpr(ops[0], op_dtypes[0], op_dtypes[1])

        res = f'({ops[0]}) {ir.OPMAP[node.operator]} ({ops[1]})'

        if node.operator in [ir.opc.RShift]:
            res = self.cast_svexpr(res, op_dtypes[0], node.dtype)

        return res

    def visit_SubscriptExpr(self, node):
        prev_ctx = self.ctx
        self.ctx = node.ctx
        val = self.visit(node.val)
        self.ctx = prev_ctx

        if val is None:
            return None

        if isinstance(node.index, ir.ResExpr):
            index = node.index.val

            index = node.val.dtype.index_norm(index)[0]

            if isinstance(index, slice):
                stop = int(index.stop) - 1
                start = int(index.start)

                if isinstance(node.val, (ir.Name, ir.AttrExpr)):
                    return f'{val}[{stop}:{start}]'
            else:
                start = index
                stop = start
                index = int(index)

                if isinstance(node.val, (ir.Name, ir.AttrExpr, ir.Component)):
                    if typeof(node.val.dtype, (Tuple, Union, Queue)):
                        return f'{val}{self.separator}{node.val.dtype.fields[index]}'
                    else:
                        return f'{val}[{index}]'

            if isinstance(node.val, ir.ResExpr):
                if typeof(node.val.dtype, Array):
                    return f'{val}[{index}]'
                elif typeof(node.val.dtype, Integral):
                    return f'{val}[{index}]'
                elif typeof(node.val.dtype, (Tuple, Union, Queue)):
                    return f'{val}{self.separator}{node.val.dtype.fields[index]}'
            else:
                start = sum(node.val.dtype[i].width for i in range(start))
                stop = sum(node.val.dtype[i].width for i in range(stop + 1)) - 1

                fname = get_slice_func(self.aux_funcs, start, stop, node.val.dtype.width,
                                       getattr(node.dtype, 'signed', False))
                return f'{fname}({val})'

        if typeof(node.val.dtype, (Array, Queue, Integer, Tuple, Union)):
            ind = self.visit(node.index)
            if isinstance(node.val, ir.Name):
                return f'{val}[{ind}]'
            else:
                fname = get_index_func(self.aux_funcs, node.val.dtype.width, node.index.dtype.width,
                                       node.dtype.width, getattr(node.dtype, 'signed', False))

                return f'{fname}({val}, {ind})'

        breakpoint()
        raise Exception('Unsupported slicing')

    def visit_ConditionalExpr(self, node):
        cond = self.visit(node.cond)
        ops = [self.visit(op) for op in node.operands]
        return f'(({cond}) ? ({ops[0]}) : ({ops[1]}))'

    def _parse_intf(self, node, context=None):
        if context is None:
            context = getattr(node, 'context', None)

        if context:
            if context == 'eot':
                return f'&{node.name}_s{self.separator}{context}'

            return f'{node.name}{self.separator}{context}'

        return f'{node.name}_s'

    def generic_visit(self, node):
        return node


def svexpr(expr, aux_funcs=None):
    sv_visit = SVExpressionVisitor(aux_funcs)
    return sv_visit.visit(expr)



import typing
import ast
from ...base_resolver import ResolverBase, ResolverTypeError
from ..svcompile import compile_gear
from pygears.hls.ast.utils import is_func_empty
from pygears.util.fileio import save_file
from pygears.conf import inject, Inject


class HLSResolver(ResolverBase):
    @inject
    def __init__(self, node):
        self.node = node
        self.generated = False

        if self.empty_impl:
            raise ResolverTypeError

        self._files = [self.file_basename]

    @property
    def empty_impl(self):
        # TODO: We are doing AST generation at least twice (once inside this
        # func, second time when HLS is run), performance impact?
        return is_func_empty(self.node.func)

    @property
    def params(self):
        return {}

    @property
    def module_name(self) -> str:
        return self.hier_path_name

    @property
    def file_basename(self):
        return f'{self.module_name}.{self.lang}'

    @property
    def files(self) -> typing.List[str]:
        return self._files

    def generate(self, template_env, outdir):
        attrib = self.cfg.get('attrib', None)
        if isinstance(attrib, str):
            attrib = [attrib]

        contents, subsvmods = compile_gear(self.node,
                                           template_env,
                                           self.module_name,
                                           outdir,
                                           attrib=attrib)

        save_file(self.file_basename, outdir, contents)

        for s in subsvmods:
            self._files.extend(s.files)



from pygears import gear, module, GearDone, Intf, alternative
from pygears.sim import delta, clk
from pygears.typing import Unit, Bool


@gear
async def reqbc_din(din) -> None:
    await clk()


@gear
async def reqbc_avail(rd: Unit) -> Bool:
    await clk()


@gear
async def reqbc_dout(rd: Unit, *, t) -> b't':
    await clk()


@gear(hdl={'impl': 'reqbc', 'hierarchical': False})
def reqbc(din, *rd: Unit) -> b'(din,)*len(rd)':
    din | reqbc_din

    dout = []
    for r in rd:
        dout.append(reqbc_dout(r, t=din.dtype))

    if len(dout) > 1:
        return tuple(dout)
    else:
        return dout[0]


def make_reqbc(din, *, ports):
    rd = [Intf(Unit) for _ in range(ports)]

    dout = reqbc(din, *rd)

    if not isinstance(dout, tuple):
        dout = (dout, )

    p = []
    for i in range(0, ports):
        p.append([rd[i], dout[i]])

    if len(p) > 1:
        return tuple(p)
    else:
        return p[0]



from pygears.typing.base import is_type
import subprocess
from pygears import PluginBase, reg, find
from pygears.core.graph import get_consumer_tree
from pygears.core.port import OutPort
from pygears.sim import log, timestep, SimPlugin
from pygears.sim.sim_gear import SimGear
from pygears.typing import typeof, TLM, Float, Uint
from pygears.typing.visitor import TypingVisitorBase
from vcd import VCDWriter
from pygears.core.hier_node import HierVisitorBase
from .sim_extend import SimExtend
import os
import fnmatch
import itertools
import atexit
from pygears.conf import inject, Inject


def match(val, include_pattern):
    def check(pattern):
        if isinstance(pattern, str):
            return fnmatch.fnmatch(val.name, pattern)
        else:
            return pattern(val)

    return any(check(p) for p in include_pattern)


class VCDTypeVisitor(TypingVisitorBase):
    def __init__(self, max_level):
        self.fields = {}
        self.max_level = max_level
        self.level = 0

    def visit(self, type_, field=None, **kwds):
        self.level += 1
        if self.level == self.max_level:
            try:
                self.visit_Uint(Uint[type_.width], field)
            except:
                pass
        else:
            super().visit(type_, field, **kwds)

        self.level -= 1

    def visit_Int(self, type_, field):
        self.fields[field] = type_

    def visit_Bool(self, type_, field):
        self.fields[field] = type_

    def visit_Uint(self, type_, field):
        self.fields[field] = type_

    def visit_Fixp(self, type_, field):
        self.fields[field] = type_

    def visit_Float(self, type_, field):
        self.fields[field] = type_

    def visit_Ufixp(self, type_, field):
        self.fields[field] = type_

    def visit_Unit(self, type_, field):
        self.fields[field] = type_

    def visit_Queue(self, type_, field):
        self.visit(type_[0], f'{field}.data' if field else 'data')
        self.visit(type_[1:], f'{field}.eot' if field else 'eot')

    def visit_Union(self, type_, field):
        self.visit(type_[0], f'{field}.data' if field else 'data')
        self.visit(type_[1], f'{field}.ctrl' if field else 'ctrl')

    def visit_Array(self, type_, field):
        for i, t in enumerate(type_):
            self.visit(t, f'{field}({i})' if field else f'({i})')

    def visit_default(self, type_, field):
        if hasattr(type_, 'fields'):
            for t, f in zip(type_, type_.fields):
                self.visit(t, f'{field}.{f}' if field else f)
        else:
            super().visit_default(type_, field)


class VCDValVisitor(TypingVisitorBase):
    def __init__(self, vcd_vars, writer, timestep, max_level):
        self.vcd_vars = vcd_vars
        self.writer = writer
        self.max_level = max_level
        self.level = 0
        self.timestep = timestep

    def visit(self, type_, field=None, val=None):
        self.level += 1
        if self.level == self.max_level:
            try:
                self.Uint(Uint[type_.width], field, val.code())
            except:
                pass
        else:
            getattr(self, type_._base.__name__)(type_, field, val=val)

        self.level -= 1

    def change(self, dtype, field, val):
        self.writer.change(self.vcd_vars[field], self.timestep, dtype(val).code())

    def Float(self, type_, field, val=None):
        self.writer.change(self.vcd_vars[field], self.timestep, float(val))

    def Union(self, type_, field, val=None):
        f0 = getattr(self, type_[0]._base.__name__)
        f1 = getattr(self, type_[1]._base.__name__)
        if field:
            f0(type_[0], f'{field}.data', val=val[0])
            f1(type_[1], f'{field}.ctrl', val=val[1])
        else:
            f0(type_[0], 'data', val=val[0])
            f1(type_[1], 'ctrl', val=val[1])

    def Queue(self, type_, field, val=None):
        f0 = getattr(self, type_.data._base.__name__)
        f1 = getattr(self, type_.eot._base.__name__)
        if field:
            f0(type_.data, f'{field}.data', val=val[0])
            f1(type_.eot, f'{field}.eot', val=val[1])
        else:
            f0(type_.data, 'data', val=val[0])
            f1(type_.eot, 'eot', val=val[1])

    def Array(self, type_, field, val):
        t = type_.dtype
        f = getattr(self, t._base.__name__)
        if field:
            for i in range(len(type_)):
                f(t, f'{field}({i})', val=val[i])
        else:
            for i in range(len(type_)):
                f(t, f'({i})', val=val[i])

    Maybe = Union
    Ufixp = change
    Fixp = change
    Uint = change
    Int = change
    Unit = change
    Bool = change

    def Tuple(self, type_, field, val):
        if field:
            for i in range(len(type_)):
                f = getattr(self, type_[i]._base.__name__)
                f(type_[i], f'{field}.{type_.fields[i]}', val=val[i])
        else:
            for i in range(len(type_)):
                f = getattr(self, type_[i]._base.__name__)
                f(type_[i], type_.fields[i], val=val[i])

    # def default(self, type_, field, val=None):
    #     if hasattr(type_, 'fields'):
    #         breakpoint()
    #         for t, f, v in zip(type_, type_.fields, val):
    #             self.visit(t, f'{field}.{f}' if field else f, val=v)


def register_traces_for_intf(dtype, scope, writer, expand_data=True):
    # TODO: Refactor this into a class
    vcd_vars = {'srcs': [], 'srcs_active': [], 'dtype': dtype}

    if typeof(dtype, TLM) or not is_type(dtype):
        vcd_vars['data'] = writer.register_var(scope, 'data', 'string')
    else:
        if not expand_data:
            try:
                vcd_vars['data'] = writer.register_var(scope,
                                                       'data',
                                                       var_type='wire',
                                                       size=max(dtype.width, 1))
            except:
                vcd_vars['data'] = writer.register_var(scope, 'data', 'string')
        else:
            v = VCDTypeVisitor(max_level=10)
            v.visit(dtype, 'data')

            for name, t in v.fields.items():
                field_scope, _, basename = name.rpartition('.')
                if field_scope:
                    field_scope = '.'.join((scope, field_scope))
                else:
                    field_scope = scope

                if typeof(t, Float):
                    vcd_vars[name] = writer.register_var(field_scope,
                                                         basename,
                                                         var_type='real',
                                                         size=32)
                else:
                    vcd_vars[name] = writer.register_var(field_scope,
                                                         basename,
                                                         var_type='wire',
                                                         size=max(t.width, 1))

    for sig in ('valid', 'ready'):
        vcd_vars[sig] = writer.register_var(scope, sig, 'wire', size=1, init=0)

    return vcd_vars


def is_trace_included(port, include, vcd_tlm):
    # if not match(f'{port.gear.name}.{port.basename}', include):
    if not match(port, include):
        return False

    if (port.dtype is None) or (typeof(port.dtype, TLM) and not vcd_tlm):
        return False

    return True


class VCDHierVisitor(HierVisitorBase):
    @inject
    def __init__(self, include, vcd_tlm, sim_map=Inject('sim/map')):
        self.include = include
        self.vcd_tlm = vcd_tlm
        self.sim_map = sim_map
        self.vcd_vars = {}
        self.end_consumers = {}
        self.indent = 0

    def enter_hier(self, name):
        self.indent += 4

    def exit_hier(self, name):
        self.indent -= 4

    def trace_if_included(self, p):
        if not is_trace_included(p, self.include, self.vcd_tlm):
            return

        gear_vcd_scope = p.gear.name[1:].replace('/', '.')

        scope = '.'.join([gear_vcd_scope, p.basename])

        self.vcd_vars[p] = scope

    def Gear(self, module):
        if module.parent is None:
            return super().HierNode(module)

        self.enter_hier(module.basename)

        for p in module.in_ports:
            self.trace_if_included(p)

        if module in self.sim_map or module.hierarchical:
            for p in module.out_ports:
                self.trace_if_included(p)

        if module in self.sim_map:
            for p in module.in_ports:
                # TODO Hack to make end_consumer unique with id(intf) so that
                # it can be looked upon in the list. Make it in a better way
                self.end_consumers[p.consumer] = {'prods': [], 'intf': id(p.consumer)}

        if (module in self.sim_map or module.hierarchical) and module.params['sim_cls'] is None:
            super().HierNode(module)

        self.exit_hier(module.basename)

        return True


class VCD(SimExtend):
    @inject
    def __init__(
            self,
            trace_fn='pygears.vcd',
            include=Inject('debug/trace'),
            tlm=False,
            shmidcat=Inject('sim_extens/vcd/shmidcat'),
            vcd_fifo=Inject('sim_extens/vcd/vcd_fifo'),
            sim=Inject('sim/simulator'),
            outdir=Inject('results-dir'),
            expand_data=Inject('debug/expand_trace_data'),
    ):
        super().__init__()
        self.sim = sim
        self.expand_data = expand_data
        self.finished = False
        self.vcd_fifo = vcd_fifo
        self.shmidcat = shmidcat
        self.outdir = outdir
        self.trace_fn = None
        self.shmid_proc = None
        self.include = include

        self.trace_fn = os.path.abspath(os.path.join(self.outdir, trace_fn))
        atexit.register(self.finish)

        try:
            subprocess.call(f"rm -f {self.trace_fn}", shell=True)
        except OSError:
            pass

        if self.vcd_fifo:
            subprocess.call(f"mkfifo {self.trace_fn}", shell=True)
        else:
            log.info(f'Main VCD dump to "{self.trace_fn}"')

        if self.shmidcat:
            self.shmid_proc = subprocess.Popen(f'shmidcat {self.trace_fn}',
                                               shell=True,
                                               stdout=subprocess.PIPE)

            # Wait for shmidcat to actually open the pipe, which is necessary
            # to happen prior to init of the verilator. If shmidcat does not
            # open the pipe, verilator will get stuck
            import time
            time.sleep(0.1)

        self.vcd_file = open(self.trace_fn, 'w')

        if self.shmidcat:
            self.shmid = self.shmid_proc.stdout.readline().decode().strip()
            log.info(f'Main VCD dump to shared memory at 0x{self.shmid}')

        self.writer = VCDWriter(self.vcd_file, timescale='1 ns', date='today')

        reg['VCDWriter'] = self.writer
        reg['VCD'] = self

        self.clk_var = self.writer.register_var('', 'clk', 'wire', size=1, init=1)

        self.timestep_var = self.writer.register_var('', 'timestep', 'integer', init=0)

        self.handhake = set()

    def before_run(self, sim):
        vcd_visitor = VCDHierVisitor(self.include, False)
        vcd_visitor.visit(find('/'))

        if not vcd_visitor.vcd_vars:
            self.deactivate('before_run')
            return True

        self.vcd_vars = {
            p: register_traces_for_intf(p.dtype, scope, self.writer, self.expand_data)
            for p, scope in vcd_visitor.vcd_vars.items()
        }

        self.end_consumers = vcd_visitor.end_consumers

        self.writer.flush()

        for intf in self.end_consumers:
            intf.events['put'].append(self.intf_put)
            intf.events['ack'].append(self.intf_ack)

        vcd_intf_vars = {}
        for p, v in self.vcd_vars.items():
            intf.events['put'].append(self.intf_put)
            intf.events['ack'].append(self.intf_ack)
            vcd_intf_vars[p] = v

        self.vcd_vars = vcd_intf_vars
        self.extend_intfs()

    def extend_intfs(self):
        for p, v in self.vcd_vars.items():
            v['srcs'] = [self.end_consumers[pp.consumer] for pp in get_consumer_tree(p.consumer)]
            v['srcs_active'] = [False] * len(v['srcs'])
            v['p'] = p
            for vs in v['srcs']:
                vs['prods'].append(v)

        reg['graph/consumer_tree'] = {}
        reg['graph/end_producer'] = {}

    def var_put(self, v, val):
        cur_timestep = timestep() * 10
        if typeof(v['dtype'], TLM):
            self.writer.change(v['data'], cur_timestep, str(val))
        else:
            try:
                if self.expand_data:
                    visitor = VCDValVisitor(v, self.writer, cur_timestep, max_level=10)
                    visitor.visit(v['dtype'], 'data', val=val)
                else:
                    self.writer.change(v['data'], cur_timestep, val.code())
            except AttributeError:
                pass

        self.writer.change(v['valid'], cur_timestep, 1)

    def intf_put(self, intf, val):
        p = intf.producer
        if p in self.vcd_vars:
            v = self.vcd_vars[p]
            self.var_put(v, val)

        if intf in self.end_consumers:
            v = self.end_consumers[intf]
            for vp in v['prods']:
                if not any(vp['srcs_active']):
                    # TODO: Optimization possibility, don't write the data, only ready/valid signals
                    self.var_put(vp, val)

                for i, vv in enumerate(vp['srcs']):
                    if vv is v:
                        vp['srcs_active'][i] = True
                        break

        return True

    def intf_ack(self, intf):
        p = intf.producer
        if p in self.vcd_vars:
            v = self.vcd_vars[p]
            self.writer.change(v['ready'], timestep() * 10, 1)
            self.handhake.add(p)

        if intf in self.end_consumers:
            v = self.end_consumers[intf]
            for vp in v['prods']:

                for i, vv in enumerate(vp['srcs']):
                    if vv is v:
                        vp['srcs_active'][i] = False
                        break

                if not any(vp['srcs_active']):
                    self.writer.change(vp['ready'], timestep() * 10, 1)
                    self.handhake.add(vp['p'])

        return True

    def before_timestep(self, sim, timestep):
        self.writer.change(self.clk_var, timestep * 10 + 5, 0)
        return True

    def after_timestep(self, sim, timestep):
        timestep += 1
        self.writer.change(self.timestep_var, timestep * 10, timestep)
        self.writer.change(self.clk_var, timestep * 10, 1)
        for p, v in self.vcd_vars.items():
            if p in self.handhake:
                self.writer.change(v['ready'], timestep * 10, 0)
                if not any(v['srcs_active']):
                    self.writer.change(v['valid'], timestep * 10, 0)

                self.handhake.remove(p)

        self.writer.flush()

        return True

    def finish(self):
        if not self.finished:
            self.writer.close()
            self.vcd_file.close()
            self.finished = True

            if self.shmid_proc:
                self.shmid_proc.terminate()

    def after_cleanup(self, sim):
        self.finish()


class SimVCDPlugin(SimPlugin):
    @classmethod
    def bind(cls):
        reg['sim_extens/vcd/shmidcat'] = False
        reg['sim_extens/vcd/vcd_fifo'] = False
        reg['sim/extens'].append(VCD)



import collections
import copy
import functools
import copyreg
import operator



class BlackBox:
    """All BlackBoxes are the same."""
    def __init__(self, contents):
        # TODO: use a weak reference for contents
        self._contents = contents

    @property
    def contents(self):
        return self._contents

    def __eq__(self, other):
        return isinstance(other, type(self))

    def __hash__(self):
        return hash(type(self))


class hashabledict(dict):
    def __hash__(self):
        return hash(tuple(self.items()))

@functools.lru_cache(maxsize=None)
def index_norm_hashable_single(i, dtype):
    size = len(dtype)
    if isinstance(i, tuple):
        start, stop, step = i

        if step == -1:
            start, stop = stop, start
            if stop is not None:
                if stop == -1:
                    stop = None
                else:
                    stop += 1

            step = 1

        if start is None:
            start = 0
        elif start < 0:
            start += size

        if stop is None:
            stop = size
        elif stop < 0:
            stop += size
        elif stop > size:
            stop = size

        # if start == stop:
        #     raise IndexError

        return slice(start, stop, step)

    elif isinstance(i, str):
        return dtype.fields.index(i)
    else:

        if i < 0:
            i = size + i

        if i >= size:
            raise IndexError(f'index {i} out of bounds')

        return i


@functools.lru_cache(maxsize=None)
def index_norm_hashable(index, dtype):
    return tuple(index_norm_hashable_single(i, dtype) for i in index)


class TemplateArgumentsError(Exception):
    pass


class TemplatedTypeUnspecified(Exception):
    pass


class TypingMeta(type):
    """Base class all types.
    """
    @property
    def specified(self):
        return True

    def __repr__(self):
        return self.__name__

    # def __eq__(self, other):
    #     return self is other

    def __hash__(self):
        return hash(self.__name__)

    def copy(self):
        return self


def type_repr(obj):
    if isinstance(obj, type) and not isinstance(obj, GenericMeta):
        if obj.__module__ == 'builtins':
            return obj.__qualname__
    if obj is ...:
        return ('...')
    return repr(obj)


def type_str(obj):
    if isinstance(obj, type) and not isinstance(obj, GenericMeta):
        if obj.__module__ == 'builtins':
            return obj.__qualname__
    if obj is ...:
        return ('...')
    return str(obj)


def pickle_c(c):
    if c.is_generic():
        return c.__name__

    return operator.getitem, (c.base, tuple(c.args))


class class_and_instance_method:
    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__

    def __get__(self, instance, cls=None):
        if instance is None:
            # return the metaclass method, bound to the class
            type_ = type(cls)
            return getattr(type_, self.func.__name__).__get__(cls, type_)
        return self.func.__get__(instance, cls)


class GenericMeta(TypingMeta):
    """Base class for all types that have a generic parameter.
    """
    _args = None
    _hash = None
    _base = None
    _specified = None

    @classmethod
    @functools.lru_cache(maxsize=None)
    def get_class(cls, name, bases, namespace, args):
        namespace = dict(namespace.contents)
        # TODO: dict parenthesis can be avoided and Python will parse dict
        # like structure as list of slices. Maybe try this to reduce clutter
        if isinstance(args, dict):
            namespace.update(
                {
                    '__args__': tuple(args.values()),
                    '__parameters__': tuple(args.keys())
                })
        else:
            if args is None:
                namespace['__args__'] = tuple()
            elif not isinstance(args, tuple):
                namespace['__args__'] = (args, )
            else:
                namespace['__args__'] = args

        if len(bases) <= 1:
            base = None
        else:
            base = bases[-2]

        namespace.update(
            {
                '_hash': None,
                '_base': base,
                '_specified': None,
                '_args': None
            })

        res = super().__new__(cls, name, bases, namespace)
        if base is None:
            res._base = res

        return res

    def __new__(cls, name, bases, namespace, args=None):
        # TODO: Throw error when too many args are supplied
        if ((not bases) or (not hasattr(bases[0], 'args')) or (not bases[0].args)
                or args is None):

            if isinstance(args, slice):
                args = {args.start: args.stop}

            if isinstance(args, tuple) and any(isinstance(a, slice) for a in args):
                args_dict = {}

                for i, val in enumerate(args):
                    if isinstance(val, slice):
                        args_dict[val.start] = val.stop
                    else:
                        args_dict[f'f{i}'] = val

                args = args_dict

            if isinstance(args, dict):
                args = hashabledict(args)

            return cls.get_class(name, bases, BlackBox(namespace), args)
        else:
            if not isinstance(args, dict) and not isinstance(args, tuple):
                args = (args, )

            if len(bases[0].templates) < len(args):
                raise TemplateArgumentsError(
                    "Too many arguments to the templated type: {bases[0]}")

            if isinstance(args, dict):
                for t in args:
                    if not any(t==str(tt) for tt in bases[0].templates):
                        raise TemplateArgumentsError(
                            f"Template parameter '{t}' not part of the "
                            f"templated type: {bases[0]}")

                tmpl_map = args
            else:
                tmpl_map = {name: val for name, val in zip(bases[0].templates, args)}

            return param_subs(bases[0], tmpl_map, {})

    def __init_subclass__(cls, **kwds):
        copyreg.pickle(cls, pickle_c)

    def is_generic(self):
        """Return True if no values have been supplied for the generic parameters.

        >>> Uint.is_generic()
        True

        >>> Uint['template'].is_generic()
        False
        """

        return len(self.args) == 0

    def is_abstract(self):
        return True

    def __bool__(self):
        return self.specified

    def __hash__(self):
        if self._hash is None:
            if bool(self.args):
                base = self.base
                if base.__name__ == 'Maybe':
                    base = 'Maybe'

                # self._hash = hash((base, tuple(self.args), tuple(self.fields)))
                self._hash = hash((base, tuple(self.args)))
            else:
                # TODO: Future expansion: what if there is two implementations of the type with the same name
                self._hash = hash(self.__name__)

        return self._hash

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def args_specified(self):
        try:
            if len(self.args) != len(self.__parameters__):
                return False
        except AttributeError:
            pass

        if self.args:
            for a in self.args:
                try:
                    if not a.specified:
                        return False
                except AttributeError:
                    if isinstance(a, (str, bytes, T)):
                        return False

            return True
        else:
            return False

    @property
    def specified(self):
        """Return True if all generic parameters were supplied concrete values.

        >>> Uint['template'].specified
        False

        >>> Uint[16].specified
        True
        """
        if self._specified is None:
            self._specified = self.args_specified

        return self._specified

    def __getitem__(self, params):
        # if isinstance(params, tuple):
        #     params = list(params)
        # elif not isinstance(params, dict):
        #     params = [params]

        if isinstance(params, dict):
            params = hashabledict(params)

        return self.__class__(
            self.__name__, (self, ) + self.__bases__, self.__dict__, args=params)

    @property
    def base(self):
        """Returns base generic class of the type.

        >>> assert Uint[16].base == Uint
        """
        if self._base is None:

            if len(self.__bases__) == 1:
                self._base = self
            else:
                self._base = self.__bases__[-2]

        return self._base

    @property
    def templates(self):
        """Returns a list of templated generic variables within the type. The type is
searched recursively. Each template is reported only once.

        >>> Tuple[Tuple['T1', 'T2'], 'T1'].templates
        ['T1', 'T2']
        """
        def make_unique(seq):
            seen = set()
            return [x for x in seq if x not in seen and not seen.add(x)]

        templates = []
        for a in self.args:
            if hasattr(a, 'templates'):
                a_templates = a.templates
                templates += [v for v in a_templates if v not in templates]
            else:
                if isinstance(a, (str, T)):  #and templ_var_re.search(a):
                    templates.append(a)

        return make_unique(templates)

    @property
    def args(self):
        """Returns a list of values supplied for each generic parameter.

        >>> Tuple[Uint[1], Uint[2]].args
        [Uint[1], Uint[2]]
        """

        if self._args is None:
            if hasattr(self, '__args__'):
                if hasattr(self, '__default__'):
                    plen = len(self.__parameters__)
                    alen = len(self.__args__)
                    dlen = len(self.__default__)
                    missing = plen - alen

                    if (missing == 0) or (dlen < missing):
                        self._args = self.__args__
                    else:
                        self._args = self.__args__ + self.__default__[-missing:]
                else:
                    self._args = self.__args__
            else:
                self._args = []

        return self._args

    def __repr__(self):
        if not self.args:
            return super().__repr__()
        else:
            return super().__repr__() + '[%s]' % ', '.join(
                [type_repr(a) for a in self.args])

    def __str__(self):
        if not self.args:
            return super().__repr__()
        else:
            return super().__str__() + '[%s]' % ', '.join(
                [type_str(a) for a in self.args])

    @property
    def fields(self):
        """Returns the names of the generic parameters.

        >>> Tuple[Uint[1], Uint[2]].fields
        ('f0', 'f1')

        >>> Tuple[{'u1': Uint[1], 'u2': Uint[2]}].fields
        ('u0', 'u1')
        """

        if hasattr(self, '__parameters__'):
            return self.__parameters__
        else:
            return [f'f{i}' for i in range(len(self.args))]

    def remove(self, *args):
        args = {k: v for k, v in zip(self.fields, self.args) if k not in args}

        return self.base[args]

    def rename(self, **kwds):
        args = {kwds.get(k, k): v for k, v in zip(self.fields, self.args)}

        return self.base[args]

    def subs(self, **kwds):
        args = {k: kwds.get(k, v) for k, v in zip(self.fields, self.args)}

        return self.base[args]

    def copy(self):
        if hasattr(self, '__parameters__'):
            args = {
                f: a.copy() if is_type(a) else copy.copy(a)
                for f, a in zip(self.fields, self.args)
            }
        else:
            args = tuple(a.copy() if is_type(a) else copy.copy(a) for a in self.args)

        return self.base[args]

    # @functools.lru_cache(maxsize=None)
    def _arg_eq(self, other):
        if len(self.args) != len(other.args):
            return False
        return all(s == o for s, o in zip(self.args, other.args))

    # def __eq__(self, other):
    #     if not isinstance(other, GenericMeta):
    #         return False

    #     if self.base is not other.base:
    #         return False

    #     if len(self.args) != len(other.args):
    #         return False

    #     return all(s == o for s, o in zip(self.args, other.args))


def param_subs(t, matches, namespace):
    t_orig = t

    if isinstance(t, bytes):
        t = t.decode()

    if isinstance(t, T):
        res = matches.get(t, namespace.get(t, None))
        if res is not None:
            return res

        name = t.__name__
        return matches.get(name, namespace.get(name, t))

    # Did we reach the parameter name?
    if isinstance(t, str):
        if t.isidentifier():
            return matches.get(t, namespace.get(t, t))

        err = None
        try:
            return eval(t, namespace, matches)
        except Exception as e:
            err = e

        if err:
            raise type(err)(f"{str(err)}\n - while evaluating string parameter '{t}'")

    elif isinstance(t, collections.abc.Iterable):
        return type(t)(param_subs(tt, matches, namespace) for tt in t)
    else:
        if isinstance(t, GenericMeta) and (not t.specified):
            args = tuple(
                param_subs(t.args[i], matches, namespace) for i in range(len(t.args)))

            if hasattr(t, '__parameters__'):
                args = {name: a for name, a in zip(t.__parameters__, args)}

            return t.__class__(t.__name__, t.__bases__, t.__dict__, args=args)

    return t_orig


class EnumerableGenericMeta(GenericMeta):
    """Base class for all types that are iterable.
    """
    @property
    def width(self):
        """Calculates the bit width of the type.

        >>> int(Tuple[Uint[1], Uint[2]])
        3
        """
        if self.specified:
            return sum(map(int, self))
        else:
            raise Exception(
                f"Cannot evaluate width of unspecified generic type"
                f" {type_repr(self)}")

    def __len__(self):
        """The number of elements type generates when iterated.

        >>> Uint[16])
        16
        """
        return len(self.keys())

    def keys(self):
        """Returns a list of keys that can be used for indexing the type.
        """
        return list(range(len(self.args)))

    def index_convert(self, index):
        if isinstance(index, str):
            try:
                return self.fields.index(index)
            except ValueError as e:
                raise KeyError(f'Field "{index}" not in type "{repr(self)}"')
        elif not isinstance(index, slice):
            return int(index)
        else:
            return index.__reduce__()[1]

    def index_norm(self, index):
        if not isinstance(index, tuple):
            return (index_norm_hashable_single(self.index_convert(index), self), )
        else:
            return index_norm_hashable(
                tuple(self.index_convert(i) for i in index), self)

    def items(self):
        """Generator that yields (key, element) pairs.
        """
        for k in self.keys():
            yield k, self[k]


class Any(metaclass=TypingMeta):
    """Type that can be matched to any other type.
    """

    def __new__(val):
        return val


def typeof(obj, t):
    """Check if a specific type instance is a subclass of the type.

    Args:
       obj: Concrete type instance
       t: Base type class

    """
    try:
        res = issubclass(obj, t)
        if not res and obj.base.__name__ == 'Maybe':
            if not isinstance(t, tuple):
                t = (t, )

            return any(ti.base.__name__ == 'Maybe' for ti in t)

        return res

    except TypeError:
        return False
    except AttributeError:
        return False


def is_type(obj):
    return isinstance(obj, TypingMeta)


class T:
    __slots__ = ('__name__', '__bound__')

    def __init__(self, name, bound):
        self.__bound__ = bound
        self.__name__ = name

    def __repr__(self):
        return self.__name__

    def __reduce__(self):
        return f'~{self.__name__}'

    def copy(self):
        return self



import logging
import pprint
import ast
import textwrap
import inspect

from .ast.utils import get_function_source

logger = None


def hls_log():
    global logger
    if logger is None:
        import sys
        logger = logging.getLogger('hls')
        h = logging.StreamHandler(sys.stdout)
        h.setLevel(logging.DEBUG)
        logger.addHandler(h)

    return logger


def hls_enable_debug_log():
    logger = hls_log()
    logger.setLevel(logging.DEBUG)


def hls_disable_debug_log():
    logger = hls_log()
    logger.setLevel(logging.ERROR)


def hls_debug_log_enabled():
    return hls_log().getEffectiveLevel() == logging.DEBUG


def hls_debug(msg='', title=None, indent=0):
    if not hls_debug_log_enabled():
        return None

    if title is not None:
        hls_debug_header(title)

    if isinstance(msg, dict):
        msg = pprint.pformat(msg)
    elif isinstance(msg, ast.AST):
        import astpretty
        msg = astpretty.pformat(msg)
    else:
        msg = str(msg)

    if title is not None:
        msg = textwrap.indent(msg, '    ')

    hls_log().debug(textwrap.indent(msg, ' ' * indent))


def hls_debug_header(msg=''):
    hls_debug()
    hls_debug('*' * 80)
    hls_debug('*')
    for line in msg.split('\n'):
        hls_debug('* ' + line)

    hls_debug('*')
    hls_debug('*' * 80)
    hls_debug()


def print_gear_parse_intro(gear, body_ast):
    hls_debug('*' * 80)
    hls_debug_header(f'Compiling code for the gear "{gear.name}" of the type '
                     f'"{gear.definition.__name__}"')

    fn = inspect.getfile(gear.func)
    try:
        _, ln = inspect.getsourcelines(gear.func)
    except OSError:
        ln = '-'

    hls_debug(get_function_source(gear.func),
              title=f'Parsing function "{gear.func.__name__}" from "{fn}", line {ln}')

    hls_debug_header('Function body AST')

    for stmt in body_ast.body:
        hls_debug(stmt)


def print_func_parse_intro(func, body_ast):
    hls_debug('*' * 80)
    hls_debug_header(f'Compiling code for the {func}')

    fn = inspect.getfile(func)
    try:
        _, ln = inspect.getsourcelines(func)
    except OSError:
        ln = '-'

    hls_debug(get_function_source(func), title=f'Parsing function {func} from "{fn}", line {ln}')

    hls_debug_header('Function body AST')

    for stmt in body_ast.body:
        hls_debug(stmt)



from pygears.typing import bitw, typeof
from pygears.hls import ir
import hashlib
import inspect
import re

class Signed:
    pass

class ElemNotFound(Exception):
    pass

class VGenOffsetVisitor:
    def visit(self, type_, path):
        visit_func_name = f'visit_{type_.__name__}'

        if hasattr(self, visit_func_name):
            return getattr(self, visit_func_name)(type_, path)

        else:
            return self.visit_default(type_, path)

    def visit_Array(self, type_, path):
        if not path:
            return []

        key = path.pop(0)
        up = key.start
        assert key.stop is None

        elem = slice(f'({up})*{type_.data.width}', None, type_.data.width)

        # w = f"{bitw(type_.width-1)}'h{hex(type_.data.width)[2:]}"
        # elem = slice(f"{w}*({up})", None, w)
        return [(elem, type_.data)] + self.visit(type_.data, path)

    def visit_default(self, type_, path):
        if not path or type_.width == 1 :
            return []

        if isinstance(path[0], slice):
            return [(path[0], None)]

        if len(getattr(type_, 'fields', [])) > 1:
            key = path.pop(0)
            offset = 0
            for t, f in zip(type_, type_.fields):
                if f == key:
                    return [(slice(t.width+offset-1, offset), t)] + self.visit(t, path)
                offset += t.width
            else:
                raise ElemNotFound

        return []

class VGenTypeVisitor:
    def __init__(self):
        self.keys = []
        self.offset = 0

    def visit(self, type_):
        for c in inspect.getmro(type_.__class__):
            visit_func_name = f'visit_{c.__name__}'

            if hasattr(self, visit_func_name):
                yield from getattr(self, visit_func_name)(type_)

        else:
            yield from self.visit_default(type_)

    def visit_Array(self, type_):
        pass

    def visit_default(self, type_):
        base_offset = self.offset
        if len(getattr(type_, 'fields', [])) > 1:
            cur_offset = base_offset
            for t, f in zip(type_, type_.fields):
                self.keys.append(f)
                yield from self.visit(t)
                self.keys.pop()
                cur_offset += t.width
                self.offset = cur_offset

        yield self.keys, base_offset, type_

def paren_matcher (n):
    res = r"[^\[\]]*?(?:\["*n+r"[^\[\]]*?"+r"\][^\[\]]*?)*?"*n
    return res[9:-2]

SLICED_ID = re.compile(r'\b(\w+)\b(?:' + paren_matcher(2) + r'|(?:\.\w+\b))+(\s*<?=)?')

def split_id(s):
    path = []
    cur = ''
    par = 0
    rng = []
    for c in s:
        if c == '[':
            if par == 0:
                if cur: path.append(cur)
                cur = ''
                rng = []
            else:
                cur += c

            par += 1
        elif c == ']':
            par -= 1
            if par == 0:
                if cur:
                    rng.append(cur)
                    path.append(rng)
                cur = ''
            else:
                cur += c

        elif c == ':' and par == 1:
            rng.append(cur)
            cur = ''
        elif c == '.':
            if par == 0:
                if cur: path.append(cur)
                cur = ''
            else:
                cur += c
        else:
            cur += c

    if cur: path.append(cur)

    return path

def wire_name(path):
    return hashlib.sha1(path.encode()).hexdigest()[:8]

def combine_ranges(a, b):
    def is_step(x):
        return x.stop is None and x.step is not None

    def is_rng(x):
        return x.stop is not None and x.step is None

    def is_index(x):
        return x.stop is None and x.step is None

    def sl_arith(x):
        try:
            return str(eval(x))
        except:
            return x

    if is_rng(a) and is_rng(b):
        return slice(sl_arith(f'{a.stop} + {b.start}'), sl_arith(f'{a.stop} + {b.stop}'))
    elif is_rng(a) and is_step(b):
        return slice(sl_arith(f'{a.stop} + {b.start}'), None, b.step)
    elif is_rng(a) and is_index(b):
        return slice(sl_arith(f'{a.stop} + {b.start}'), None)
    elif is_step(a) and is_rng(b):
        return slice(sl_arith(f'{a.start} + {b.stop}'), None, sl_arith(f'{b.start} - {b.stop} + 1'))
    elif is_step(a) and is_step(b):
        return slice(sl_arith(f'{a.start} + {b.stop}'), None, b.step)
    elif is_step(a) and is_index(b):
        return slice(sl_arith(f'{a.start} + {b.start}'), None)
    else:
        return a


def rewrite(module, index):
    def substitute(m):
        name = m.group(1)
        lval = m.group(2)
        if lval is None:
            lval = ''

        if name not in index:
            return m.group(0)

        hit = m.group(0)
        if lval:
            hit = hit[:-len(lval)]

        path = []
        for i, p in enumerate(split_id(hit)):
            if i == 0:
                continue

            if isinstance(p, list):
                up = SLICED_ID.sub(substitute, p[0])
                down = None
                if len(p) > 1:
                    down = SLICED_ID.sub(substitute, p[1])

                path.append(slice(up, down))
            else:
                path.append(p)

        dtype = index[name]
        if typeof(dtype, ir.IntfType):
            if path[0] in ['valid', 'ready', 'data']:
                name = name + '_' + path[0]
                spath = [(p, None) for p in path[1:]]
            else:
                print(module)
                print(m.group(0))
                print(name)
                print(index)
                breakpoint()
                raise Exception

        else:
            try:
                spath = VGenOffsetVisitor().visit(dtype, path)
            except ElemNotFound:
                return m.group(0)

        s = name
        if spath:
            rng, dtype = spath[0]
            for p, dtype in spath[1:]:
                rng = combine_ranges(rng, p)

            if rng.start is None:
                breakpoint()

            if rng.stop is not None:
                s += f'[{rng.start}:{rng.stop}]'
            elif rng.step is not None:
                s += f'[{rng.start}+:{rng.step}]'
            else:
                s += f'[{rng.start}]'

            if getattr(dtype, 'signed', False) and not lval:
                s = f'$signed({s})'

        return s+lval

    module = SLICED_ID.sub(substitute, module)

    pat = re.compile(r'always_ff @\(posedge (\w+)\)')
    module = pat.sub(lambda m: f'always @(posedge {m.group(1)})', module)

    pat = re.compile(r'always_comb')
    module = pat.sub(lambda m: 'always @*', module)

    return module

# from pygears.typing import Tuple, Uint
# ret = rewrite('din_s.a = ~din.valid+din_s.b[0]-din_s.c[0:0]+din_s.ba', {'din': Intf(Tuple['a': Uint[2], 'b': Uint[3], 'c': Uint[1]])})
# print(ret)

# from pygears.typing import Queue, Uint
# ret = rewrite('eot_reg <= d.eot', {'d': Queue[Uint[2]]})
# print(ret)

# from pygears.typing import Tuple, Uint, Array, Int
# intfs = {
#     'din_s':
#     Tuple['a':Array[Uint[2], 4], 'b':Array[Tuple['c':Int[3], 'd':Uint[4]], 2]]
# }

# ret = rewrite('din_s.b[i] = din_s.b[2].d-din_s.b[0].c[1:0]', intfs)

# print(ret)
# breakpoint()
# print(ret[1])
# print(ret[0])

# print(split_id('din_s[qrange_stop_dout_s.data[2:1]:0]'))

# from pygears.typing import Tuple, Uint, Array, Int, Queue
# intfs = {'din': Intf(Array[Uint[16], 4]),
#          'din_s': Array[Uint[16], 4],
#          'qrange_stop_dout': Intf(Queue[Uint[3], 1]),
#          'qrange_stop_dout_s': Queue[Uint[3], 1]}


# ret = rewrite('din_s[qrange_stop_dout_s.data[2]]', intfs)
# ret = rewrite('qrange_stop_dout_s.data[2:1]', intfs)
# print(paren_matcher(2))
# import parser

# pat = re.compile(r'(?:\[[^\[\]]*?(?:\[[^\[\]]*?\][^\[\]]*?)*?\][^\[\]]*?)')
# pat = re.compile(paren_matcher(2))
# # res = pat.findall('din_s(qrange_stop_dout_s.data(2:1):0)')
# print(paren_matcher(2))
# res = pat.findall('din_s[qrange_stop_dout_s.data[2:1]:0]')
# print(res)
# ret = parser.expr('din_s[qrange_stop_dout_s.data[2:1]:0]')
# breakpoint()



import random

from pygears import gear
from pygears.core.util import perpetum
from pygears.sim import clk


@gear
async def delay_gen(din, *, f) -> b'din':
    """Delays the data by waiting a given number of clock cycles

    Args:
        f: iterable which specifies the delay values
    """
    async with din as item:
        try:
            for i in range(next(f)):
                await clk()
        except StopIteration:
            pass

        yield item


def delay(cycles):
    return delay_gen(f=perpetum(lambda x: x, cycles))


def delay_rng(start, stop):
    """Delays each input data for a random number of clock cycles which is chosen
    from a (start, stop) range.
    """
    return delay_gen(f=perpetum(random.randint, start, stop))



import logging
import inspect
import os

from pygears import Intf
from pygears.conf import PluginBase, register_custom_log, reg
from pygears.core.gear import GearPlugin


def register_hdl_paths(*paths):
    for p in paths:
        reg['hdl/include'].append(
            os.path.abspath(os.path.expandvars(os.path.expanduser(p))))


def rename_ambiguous(modname, lang):
    if (modname, lang) in reg['hdlgen/disambig']:
        return f'{modname}_{lang}'

    return modname


def hdl_log():
    return logging.getLogger('hdlgen')


def mod_lang(module):
    if module is None:
        return reg['hdl/lang']

    if isinstance(module, Intf):
        lang = None
    else:
        lang = module.meta_kwds.get('hdl', {}).get('lang', None)

    if lang is not None:
        return lang

    hdl_top = reg['hdl/top']

    if module is hdl_top:
        return reg['hdl/lang']

    if hdl_top and module is hdl_top.parent:
        return reg['hdl/toplang']

    # # TODO: We shouldn't need this?
    if module.parent is None:
        return reg['hdl/lang']

    return mod_lang(module.parent)


def get_hdlgen_cls(module, lang):
    hdlgen_cls = module.meta_kwds.get('hdl', {}).get('hdlgen_cls', None)
    if hdlgen_cls is not None:
        return hdlgen_cls

    namespace = reg[f'{lang}gen/module_namespace']

    hdlgen_cls = namespace.get(module.definition, None)
    if hdlgen_cls is not None:
        return hdlgen_cls

    for base_class in inspect.getmro(module.__class__):
        if base_class.__name__ in namespace:
            return namespace[base_class.__name__]


def hdlmod(module, lang=None):
    if lang is None:
        lang = mod_lang(module)

    hdlgen_map = reg[f'hdlgen/map']
    if module in hdlgen_map:
        return hdlgen_map[module]

    hdlgen_cls = get_hdlgen_cls(module, lang)

    if hdlgen_cls:
        inst = hdlgen_cls(module)
    else:
        inst = None

    hdlgen_map[module] = inst

    return inst


class HDLPlugin(GearPlugin):
    @classmethod
    def bind(cls):
        register_custom_log('hdl', logging.WARNING)
        reg['gear/params/meta'].subreg('hdl')
        reg['gear/params/extra/__hdl__'] = None

        reg.confdef('hdl/include', default=[])
        reg.confdef('hdl/lang', default='sv')
        reg.confdef('hdl/toplang', default=None)
        reg['hdl/top'] = None

        reg.confdef('debug/hide_interm_vals', default=True)
        reg.confdef('debug/expand_trace_data', default=True)
        reg.confdef('debug/trace_end_cycle_dump', default=False)


from .util import flow_visitor, HDLGearHierVisitor
from .common import list_hdl_files
from . import sv
from . import v
from .hdlgen import hdlgen
from .ipgen import ipgen
from .synth import synth
from pygears.sim.extens import websim

__all__ = ['hdlgen', 'list_hdl_files', 'flow_visitor']



import os
import jinja2
from .axi import get_axi_conf
from pygears import reg, find
from pygears.typing import Queue, typeof
from pygears.core.gear import InSig
from pygears.hdl import hdlgen
from pygears.hdl.templenv import get_port_intfs
from pygears.hdl.templenv import TemplateEnv
from pygears.typing.math import ceil_chunk, ceil_div, ceil_pow2, bitw
from pygears.util.fileio import save_file
from . import axi_intfs


def generate(top, intfdef, rst=True):
    files = set()
    if isinstance(top, str):
        top = find(top)

    modinst = reg['hdlgen/map'][top]

    sigs = []
    for s in top.signals.values():
        if s.name == 'clk':
            sigs.append(InSig('aclk', 1))
        elif s.name == 'rst':
            sigs.append(InSig('aresetn', 1))
        else:
            sigs.append(s)

    intfs = {p['name']: p for p in get_port_intfs(top)}

    for i in intfs.values():
        dtype = i['type']
        w_data = i['width']
        w_eot = 0
        if typeof(dtype, Queue):
            w_data = dtype.data.width
            w_eot = dtype.eot.width

        i['w_data'] = w_data
        i['w_eot'] = w_eot

    defs = []
    for name, p in intfdef.items():
        if p.t == 'axidma':
            params = {n: c.params for n, c in p.comp.items()}
            defs.extend(axi_intfs.port_def(axi_intfs.AXI_MASTER, name, **params))

            params = {n: c.params for n, c in intfdef[f'{name}_ctrl'].comp.items()}
            defs.extend(axi_intfs.port_def(axi_intfs.AXIL_SLAVE, f'{name}_ctrl', **params))

            files.update({'sfifo.v', 'axi_addr.v', 'skidbuffer.v'})

            if 'rdata' in p.comp:
                files.add('aximm2s.v')

            if 'wdata' in p.comp:
                files.add('axis2mm.v')

        elif p.t in ['bram', 'bram.req', 'axi']:

            params = {n: c.params for n, c in p.comp.items()}

            pdefs = axi_intfs.port_def(axi_intfs.AXI_SLAVE, name, **params)

            if 'rdata' in params:
                files.add('axi_slave_read.v')

            if 'wdata' in params:
                files.add('axi_slave_write.v')

            if 'rdata' in params or 'wdata' in params:
                files.update({'sfifo.v', 'axi_addr.v', 'skidbuffer.v'})

            defs.extend(pdefs)

        elif p.t == 'axis':
            if p.direction == 'w':
                tmplt = axi_intfs.AXIS_SLAVE
            else:
                tmplt = axi_intfs.AXIS_MASTER

            params = {n: c.params for n, c in p.comp.items()}

            pdefs = axi_intfs.port_def(tmplt, name, **params)

            defs.extend(pdefs)

    context = {
        'wrap_module_name': f'wrap_{modinst.module_name}',
        'module_name': modinst.wrap_module_name,
        'inst_name': modinst.wrap_module_name,
        'intfs': intfs,
        'sigs': sigs,
        'rst': rst,
        'param_map': modinst.params if not modinst.wrapped else {},
        'port_def': defs,
        'ports': intfdef
    }

    context['pg_clk'] = 'aclk'
    tmplt = 'wrapper.j2'

    base_addr = os.path.dirname(__file__)
    lang_dir = os.path.join(os.path.dirname(base_addr), 'sv')
    env = TemplateEnv(lang_dir)

    env.jenv.globals.update(
        zip=zip,
        ceil_pow2=ceil_pow2,
        ceil_div=ceil_div,
        bitw=bitw,
        ceil_chunk=ceil_chunk,
        axi_intfs=axi_intfs)

    return env.render(base_addr, tmplt, context), files



from pygears import gear
from pygears.typing import is_type
from .dreg import dreg
from .decouple import decouple


@gear
def pipeline(din, *, length, feedback=False, init=None) -> b'din':
    if init is None:
        init = []

    if is_type(init) or not isinstance(init, (list, tuple)):
        init = [init] * length

    if feedback:
        if init:
            stage_init = init[0]
            init.pop(0)
        else:
            stage_init = None

        din = decouple(din, init=stage_init)
        length -= 1

    for i in range(length):
        if init:
            stage_init = init[0]
            init.pop(0)
        else:
            stage_init = None

        din = dreg(din, init=stage_init)

    return din



import collections
import functools
import inspect
import itertools
import pickle
import time
import os
from enum import IntEnum

from pygears.core.port import HDLProducer
from pygears.typing import Integer, Uint
from pygears.lib.verif import TypeMonitorVisitor, Partial
from pygears.util.utils import quiter


class CoverError(Exception):
    pass


class CoverBin:
    '''Holds count for specific value or values defined by enablement'''

    def __init__(self, name, threshold=None, enablement=lambda x: True):
        self.name = name
        self.threshold = threshold
        self.enablement = enablement
        self.cover_cnt = 0

    def sample(self, val):
        if self.enablement(val):
            self.cover_cnt += 1

    @property
    def covered(self):
        if self.threshold:
            return self.cover_cnt >= self.threshold
        else:
            return self.cover_cnt > 0

    @property
    def hit_percent(self):
        if self.covered:
            return 100
        else:
            if self.threshold:
                return 100 * (self.cover_cnt) / self.threshold
            else:
                return 0


class CoverPoint:
    '''Cover point for one type.
    Bins can be set as list of CoverBins or left as None
    '''

    def __init__(self,
                 name,
                 dtype=None,
                 bins=None,
                 ignore_bins=None,
                 threshold=None,
                 bind_field_name=None,
                 bind_dtype=False,
                 auto_bin_max=64):
        self.name = name
        self.dtype = dtype
        self.bind_field_name = bind_field_name
        self.bind_dtype = bind_dtype
        self.auto_bin_max = auto_bin_max
        self.ignore_bins = ignore_bins if ignore_bins else []
        self.threshold = threshold
        if bins:
            self.bins = bins
            self.set_default_bin()
        else:
            self.bins = self.set_auto_bin()

        # set sub threshold only if not already set
        for b in self.bins:
            if not b.threshold:
                b.threshold = self.threshold

    def set_default_bin(self):
        '''Bin for holding all values which are not covered by other bins'''
        for i, b in enumerate(self.bins):
            if b.name == 'default':
                other = [
                    sub.enablement for j, sub in enumerate(self.bins) if j != i
                ]
                other.extend([
                    sub.enablement for j, sub in enumerate(self.ignore_bins)
                    if j != i
                ])

                def default_enablement(val, other=other):
                    for func in other:
                        if func(val):
                            return False
                    return True

                b.enablement = default_enablement
                break

    def set_auto_bin(self):
        '''Creates automatic (explicit) bins
        The number of bins created is set by auto_bin_max and
        the enablement is if a value bellongs to the bin range
        '''
        if not self.dtype:
            raise CoverError(
                f'dtype must be set for CoverPoint if bins set as None')
        if not issubclass(self.dtype, Integer):
            raise CoverError(
                f'Automatic (implicit) bins only supported for integers')
        if self.ignore_bins:
            raise CoverError(
                'Automatic bins do not support ignore bins for now')

        bins = []
        if self.auto_bin_max < 2**len(self.dtype):
            bin_num = self.auto_bin_max
            bin_rng = 2**len(self.dtype) // self.auto_bin_max
        else:
            bin_num = 2**len(self.dtype)
            bin_rng = 1

        for i in range(bin_num):
            bins.append(
                CoverBin(
                    f'auto_bin{i}',
                    enablement=
                    (lambda y: (lambda x: (x in range(y * bin_rng, (y + 1) * bin_rng)))
                     )(i)))
        return bins

    def sample(self, val):
        for b in self.bins:
            b.sample(val)

    @property
    def cover_cnt(self):
        return sum([b.cover_cnt for b in self.bins])

    @property
    def covered(self):
        return all([b.covered for b in self.bins])

    @property
    def hit_percent(self):
        return 100 * sum([b.covered for b in self.bins]) / len(self.bins)


class CoverageTypeVisitor:
    def __init__(self, cover_points):
        self.cover_points = cover_points

    def sample(self, dtype, field, data):
        for cp in self.cover_points:
            if cp.bind_field_name:
                if cp.bind_field_name == field:
                    cp.sample(data)
            elif cp.bind_dtype:
                if cp.dtype == dtype:
                    cp.sample(data)
            elif not cp.bind_field_name and not cp.bind_dtype:
                cp.sample(data)

    def set_field_prefix(self, field):
        if field:
            return field + '.'
        else:
            return ''

    def visit(self, dtype, field=None, data=None):
        visit_func_name = f'visit_{dtype.__name__.lower()}'

        visit_func = getattr(self, visit_func_name, self.visit_default)
        ret = visit_func(dtype, field, data)
        if inspect.isgenerator(ret):
            yield from ret
        else:
            yield ret

    def visit_default(self, dtype, field=None, data=None):
        self.sample(dtype, field, data)
        yield data

    def visit_union(self, dtype, field=None, data=None):
        self.sample(dtype, field, data)
        field = self.set_field_prefix(field)
        for _ in self.visit(dtype[0], f'{field}data', data=data.data):
            pass
        for _ in self.visit(dtype[-1], f'{field}ctrl', data=data.ctrl):
            pass
        yield data

    def visit_queue(self, dtype, field=None, data=None):
        self.sample(dtype, field, data)
        field = self.set_field_prefix(field)
        for (i, d), eot in quiter(enumerate(data)):
            for ret in self.visit(dtype.sub(), f'{field}data', data=d):
                if dtype.lvl == 1:
                    self.sample(Uint[1], 'eot', data=eot)
                    yield (ret, Uint[1](eot))
                else:
                    self.sample(
                        dtype.eot,
                        'eot',
                        data=(int(ret[1]) + (eot << ret[1].width)))
                    yield (
                        ret[0],
                        Uint[ret[1].width + 1](ret[1]) + (eot << ret[1].width))

    def visit_tuple(self, dtype, field=None, data=None):
        self.sample(dtype, field, data)
        field_p = self.set_field_prefix(field)
        for i, d in enumerate(data):
            if hasattr(dtype, '__parameters__'):
                field = field_p + dtype.__parameters__[i]
            else:
                field = field_p + f'f{i}'
            t = dtype[i]
            for _ in self.visit(t, field, data=d):
                pass
        yield data

    def visit_array(self, dtype, field=None, data=None):
        self.sample(dtype, field, data)
        field = self.set_field_prefix(field)
        for i, d in enumerate(data):
            for _ in self.visit(dtype[i], f'{field}f{i}', data=d):
                pass
        yield data


class CoverGroup:
    '''Group holds all cover points for one variable
    and calls the visitor for sampling'''

    def __init__(self, name, dtype, cover_points=None):
        self.name = name
        self.dtype = dtype
        self.cover_points = cover_points

    def sample(self, val):
        for _ in CoverageTypeVisitor(cover_points=self.cover_points).visit(
                self.dtype, data=val):
            pass

    def report(self):
        r = '=' * 80 + '\n'
        r += f'Coverge report for group {self.name}\n'
        r += '=' * 80 + '\n'
        for p in self.cover_points:
            r += f'Cover point: {p.name}'
            r += f' ({p.hit_percent:.2f}%, cover_cnt: {p.cover_cnt}, threshold: {p.threshold})\n'
            r += 'Bins\n'
            for b in p.bins:
                r += f'\t{b.name}: {b.hit_percent:.2f}%, cover_cnt: {b.cover_cnt}, threshold: {b.threshold}\n'
            r += '-' * 80 + '\n'
        r += '=' * 80 + '\n'
        return r


class CoverSingleValue:
    '''Coverage of single unpacked values. Combine before sampling'''

    def __init__(self, cg=None, en=True):
        self.cg = cg
        self.en = en
        self.data = None
        self.v = TypeMonitorVisitor(self.cg.dtype)

    def sample(self, item):
        self.data = self.v.visit(self.data, item, self.cg.dtype)
        if not (isinstance(self.data, Partial) or self.data is None):
            if self.en:
                self.cg.sample(self.data)
            self.data = None


class CoverIterator(collections.abc.Iterator):
    '''Coverage of iterators.
    Wraps around the iterator and copies the values'''

    def __init__(self, it, cg=None, en=True):
        self.it, self.nextit = itertools.tee(iter(it))
        self.cg = cg
        self.en = en
        self._advance()

    def _advance(self):
        try:
            self.lookahead = next(self.nextit)
            if self.en:
                self.cg.sample(self.lookahead)
        except StopIteration:
            pass

    def __next__(self):
        self._advance()
        return next(self.it)


class FuncSampleType(IntEnum):
    result = 0
    args = 1


class CoverBase:
    def __init__(self, cg=None, en=True):
        self.cg = cg
        self.en = en
        self.cover_single = None

    def sample(self, val):
        if self.en:
            if inspect.isgenerator(val):
                return CoverIterator(it=val, cg=self.cg, en=self.en)
            else:
                if not self.cover_single:
                    self.cover_single = CoverSingleValue(
                        cg=self.cg, en=self.en)
                self.cover_single.sample(val)
        return val


class CoverFunction(CoverBase):
    '''Wrapper around functions that need to be covered'''

    def __init__(self,
                 func,
                 cg=None,
                 en=True,
                 sample_type=FuncSampleType.result):
        super(CoverFunction, self).__init__(cg, en)
        functools.update_wrapper(self, func)
        self.func = func
        self.sample_type = sample_type

    def __call__(self, *arg, **kw):
        result = self.func(*arg, **kw)
        if self.sample_type == FuncSampleType.result:
            result = self.sample(result)
        else:
            self.sample(*arg, **kw)
        return result


def cover_func(cg, en=True, sample_type=FuncSampleType.result):
    '''Decorator for functions that need to be covered'''

    def cover(func):
        return CoverFunction(func=func, cg=cg, en=en, sample_type=sample_type)

    return cover


class CoverIntf(CoverBase):
    def __init__(self, intf, cg=None, en=True):
        super(CoverIntf, self).__init__(cg, en)
        self.intf = intf

    def event(self, intf, val):
        if intf is self.intf:
            self.sample(val)
        return True


def cover_intf(intf, cg, en=True):
    '''Function to register which interface needs to be covered'''
    while not isinstance(intf.producer, HDLProducer) and intf.producer:
        intf = intf.producer

    c = CoverIntf(intf, cg=cg, en=en)
    intf.events['put'].append(c.event)


def save_coverage(cg, outdir):
    timestr = time.strftime('%Y%m%d_%H%M%S')
    del cg.dtype
    for i in range(len(cg.cover_points)):
        del cg.cover_points[i].dtype
        for j in range(len(cg.cover_points[i].bins)):
            del cg.cover_points[i].bins[j].enablement
    path = os.path.join(outdir, f'{cg.name}_{timestr}.p')
    pickle.dump(cg, open(path, 'wb'))


def merge_cover_groups(cg0, cg1):
    cg = cg0
    for cp_idx, cp in enumerate(cg.cover_points):
        for bin_idx in range(len(cp.bins)):
            cg.cover_points[cp_idx].bins[
                bin_idx].cover_cnt += cg1.cover_points[cp_idx].bins[
                    bin_idx].cover_cnt
    return cg


def merge_coverage(name, outdir):
    merged_cg = None
    for fn in os.listdir(outdir):
        if name in fn:
            curr_cg = pickle.load(open(os.path.join(outdir, fn), 'rb'))
            if merged_cg:
                merged_cg = merge_cover_groups(merged_cg, curr_cg)
            else:
                merged_cg = curr_cg
    return merged_cg



from pygears import alternative, gear
from pygears.lib import cart
from pygears.typing import Queue, Tuple, Uint, Bool


@gear
async def chop(din: Queue[Tuple[{
        'data': 't_data',
        'size': Uint
}]], *, init=1) -> Queue['t_data']:
    """Chops the input transaction into sub-transactions by sending the lower
    ``eot`` after a given number of data has passed (specified by the ``size``
    field of the :class:`Tuple`)

    Args:
        init: Initialization value for the counter

    Returns:
        A level 2 :class:`Queue` type whose data consists of the ``data`` field
          of the :class:`Tuple` input
    """

    cnt = din.dtype.data['size'](init)
    last: Bool

    async for ((data, size), eot) in din:
        last = (cnt == size)

        yield (data, eot or last)

        if last:
            cnt = init
        else:
            cnt += 1


@alternative(chop)
@gear
def chop2(din: Queue, size: Uint) -> b'din':
    return cart(size, din, order=[1, 0]) | chop



import math
import os

import jinja2

from pygears import reg
from pygears.typing import Queue, typeof, Tuple, Union, Array, Queue, Integral
from pygears.hdl.templenv import TemplateEnv
from pygears.sim.extens.rand_base import RandBase
from pygears.hdl.sv.util import svgen_typedef
from pygears.util.fileio import save_file
from pygears.conf import inject, Inject
from pygears.sim.extens.svsock import register_intf, CMD_READ


class SVRandConstraints:
    def __init__(self,
                 name='dflt',
                 cons=[],
                 dtype=None,
                 cls='dflt_tcon',
                 cls_params=None,
                 tenv=None):
        self.name = name
        self.cons = cons.copy()
        self.dtype = dtype
        self.cls = cls
        self.cls_params = cls_params
        self.cvars = {}
        self.cvars['dout'] = svgen_typedef(dtype, 'dout')
        self.tenv = tenv

    def add_var(self, name, dtype):
        self.cvars[name] = svgen_typedef(dtype, name)

    def get_class(self):
        return self.tenv.cons.default_tcon(self)


class QueueConstraints(SVRandConstraints):
    def get_class(self):
        queue_struct_cons = []
        queue_size_cons = []
        data_cons = []

        for c in self.cons:
            if 'queue.length' in c:
                queue_size_cons.append(c.replace('queue.length', 'length'))
            elif 'queue.struct' in c:
                queue_struct_cons.append(c.replace('queue.struct', 'struct'))
            else:
                data_cons.append(c.replace(f'dout.data', 'data'))

        self.cons = data_cons
        context = {
            'con': self,
            'sv_dtype': self.cvars['dout'],
            'queue_struct_cons': queue_struct_cons,
            'queue_size_cons': queue_size_cons
        }

        if typeof(self.dtype.data, Integral):
            dt = self.dtype.data
            context[
                'sv_data_dtype'] = f'logic {"signed" if dt.signed else ""} [{dt.width-1}:0]'
        else:
            context['sv_data_dtype'] = f'dout_data_t'

        del self.cvars['dout']

        return self.tenv.cons.queue_tcon(**context)


class SVServerIntf:
    def __init__(self):
        pass

    def read(self):
        return "rand_i.get_rand(synchro_handle, data[15:0]);"


class SVServerModule:
    def __init__(self, outdir, constraints, tenv):
        self.tenv = tenv
        self.outdir = outdir
        self.constraints = constraints

    def declaration(self):
        return 'svrand_top rand_i();'

    def files(self):
        files = [os.path.join(self.outdir, 'svrand_top.sv')]

        # custom classes
        for con in self.constraints:
            if con.cls == 'qenvelope':
                files.append(
                    os.path.join(self.outdir, f'qenvelope_{con.name}.sv'))

        return files


def default_tcon_resolver(desc, tenv):
    tcons = SVRandConstraints(name=desc['name'],
                              dtype=desc['dtype'],
                              cons=desc['cons'],
                              cls=desc['cls'],
                              cls_params=desc['cls_params'],
                              tenv=tenv)

    for name, dtype in desc['params'].items():
        tcons.add_var(name, dtype)

    return tcons


def queue_tcon_resolver(desc, tenv):
    tcons = QueueConstraints(name=desc['name'],
                             dtype=desc['dtype'],
                             cons=desc['cons'],
                             cls=desc['cls'],
                             cls_params=desc['cls_params'],
                             tenv=tenv)

    for name, dtype in desc['params'].items():
        tcons.add_var(name, dtype)

    return tcons


tcon_resolvers = {
    Queue: queue_tcon_resolver,
}


def find_tcon_resolver(desc, tenv):
    for templ in tcon_resolvers:
        if typeof(desc['dtype'], templ):
            return tcon_resolvers[templ](desc, tenv)

    return default_tcon_resolver(desc, tenv)


class SVRandSocket(RandBase):
    @inject
    def __init__(self, top, cons=Inject('sim/svrand/constraints')):
        basedir = os.path.dirname(__file__)
        self.tenv = TemplateEnv(basedir)
        self.tenv.cons = self.tenv.load(basedir, 'svrand_cons.j2').module

        super().__init__(top, cons)

    def create_type_cons(self, desc={}):
        return find_tcon_resolver(desc, self.tenv)

    def before_setup(self, sim):
        basedir = os.path.dirname(__file__)
        tenv = TemplateEnv(basedir)

        self.create_svrand_top()
        register_intf(SVServerModule(self.outdir, self.constraints, tenv))

        for c in self.constraints:
            c.index = register_intf(SVServerIntf())

        self.conn = reg['sim/svsock/server']
        self.send_cmd = self.conn.send_cmd
        self.dtype_recv = self.conn.dtype_recv

    def parse_name(self, name):
        for c in self.constraints:
            if c.name == name:
                return c.index, c.dtype

    def send_req(self, index, dtype):
        self.send_cmd(CMD_READ | index)
        return self.dtype_recv(dtype)

    def get_rand(self, name):
        index, dtype = self.parse_name(name)
        data = self.send_req(index, dtype)
        return data

    def create_svrand_top(self):
        base_addr = os.path.dirname(__file__)
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(base_addr),
                                 trim_blocks=True,
                                 lstrip_blocks=True)

        context = {
            'tcons': self.constraints,
        }

        res = env.get_template('svrand_top.j2').render(context)
        save_file('svrand_top.sv', self.outdir, res)

        # custom classes
        for con in self.constraints:
            if con.cls == 'qenvelope':
                context = {'tcon': con}
                res = env.get_template('qenvelope.j2').render(context)
                save_file(f'qenvelope_{con.name}.sv', self.outdir, res)



from pygears import alternative, TypeMatchError, gear
from pygears.typing import Array
from pygears.lib import fmap as common_fmap
from pygears.lib.ccat import ccat


@alternative(common_fmap)
@gear(enablement=b'typeof(din, Array)')
def arraymap(din, *, f, fcat=ccat):
    res = tuple(d | f for d in din)
    return fcat(*res) | Array[res[0].dtype, len(res)]



from pygears.sim import SimPlugin

from itertools import islice

from pygears import reg
from pygears.core.util import perpetum
from pygears.sim import log
from pygears.sim.extens.scvrand import SCVRand
from pygears.sim.extens.svrand import SVRandSocket
from pygears.typing import Queue, typeof, Bool, Tuple, Uint
from pygears.typing.queue import QueueMeta


def register_exten():
    if SVRandSocket not in reg['sim/extens']:
        reg['sim/extens'].append(SVRandSocket)
        if reg['sim/simulator'] is not None:
            SVRandSocket(top=None)


def get_rand(name, cnt=None, rand_func=None):
    if rand_func is None:
        randomizer = reg['sim/config/randomizer']
        rand_func = perpetum(randomizer.get_rand, name)

    if cnt is not None:
        yield from islice(rand_func, cnt)
    else:
        yield from rand_func


def queue_rand_seq(dtype, name):
    randomizer = reg['sim/config/randomizer']

    while True:
        val = randomizer.get_rand(name)
        yield val

        if val.eot == dtype.eot.max:
            break


def rand_seq(name, cnt=None):
    randomizer = reg['sim/config/randomizer']
    dtype = randomizer.get_dtype_by_name(name)

    if typeof(dtype, Queue):
        yield from get_rand(name, cnt, perpetum(queue_rand_seq, dtype, name))
    else:
        yield from get_rand(name, cnt)


class ConstraintWrap:
    def __init__(
            self,
            dtype,
            name,
            cons=[],
            cnt=None,
            params={},
            cls='dflt_tcon',
            cls_params=None):
        self.dtype = dtype
        self.name = name

        self.cons = cons
        self.params = params
        self.cls = cls
        self.cls_params = cls_params
        self.is_queue = False

    def get_data_desc(self):
        return {
            'name': self.name if not self.is_queue else f'{self.name}_data',
            'dtype': self.dtype if not self.is_queue else self.dtype[0],
            'cons': self.cons,
            'params': self.params,
            'cls': self.cls,
            'cls_params': self.cls_params
        }

    def get_eot_desc(self):
        return {
            'name': f'{self.name}_eot',
            'dtype': self.dtype.eot,
            'cons': self.eot_cons,
            'params': self.eot_params,
            'cls': self.eot_cls,
            'cls_params': self.eot_cls_params
        }


def randomize(
        dtype,
        name,
        cnt=None,
        cons=None,
        params=None,
        cls='dflt_tcon',
        cls_params=None):

    if cons is None:
        cons = []

    if params is None:
        params = {}

    register_exten()
    cons = ConstraintWrap(**locals())
    reg['sim/svrand/constraints'].append(cons)
    return rand_seq(name, cnt=cnt)


class SVRandPlugin(SimPlugin):
    @classmethod
    def bind(cls):
        reg['sim/svrand/constraints'] = []
        # reg['sim/svsock/server'] = None



from .cast import value_cast, type_cast
from .fixp import Fixp, Ufixp
from .uint import Uint, Bool, Int, code


def get_out_type(val_type, fract):
    if get_cut_bits(val_type, fract) <= 0:
        raise TypeError(
            f'Cannot qround type "{val_type}" with "{val_type.fract}" '
            f'fractional bits, to produce the type with more fractional '
            f'bits "fract={fract}"'
        )

    if fract != 0:
        return val_type.base[val_type.integer + 1, val_type.integer + fract + 1]
    else:
        return (Int if val_type.signed else Uint)[val_type.integer + 1]


def get_cut_bits(val_type, fract):
    return val_type.fract - fract


def qround(val, fract=0):
    cut_bits = get_cut_bits(type(val), fract)
    out_type = get_out_type(type(val), fract)
    val_coded = code(val, Int) if type(val).signed else code(val)

    res = val_coded + (Bool(1) << (cut_bits - 1))

    return out_type.decode(res[cut_bits:])


def qround_even(val, fract=0):
    cut_bits = get_cut_bits(type(val), fract)
    out_type = get_out_type(type(val), fract)
    val_coded = code(val, Int) if type(val).signed else code(val)

    round_bit = val_coded[cut_bits]

    res = val_coded + Uint([round_bit] + [~round_bit] * (cut_bits - 1))
    return out_type.decode(res[cut_bits:])



from pygears import gear
from pygears.lib.union import ucase, select
from pygears.lib.ccat import ccat
from pygears.lib.const import fix
from pygears.lib.filt import filt
from pygears.lib.fmaps.union import unionmap
from pygears.typing import Maybe, cast, Bool


@gear
def resolve(din: Maybe, *, val):
    val = din.dtype.types[1](val)
    return din | ucase(f=(fix(val=val), None))


@gear
def apply(din: Maybe, *, f, args=None, balance=None):
    maybe_args = []
    maybe_kwds = {}

    if isinstance(args, (list, tuple)):
        for a in args:
            maybe_args.append(sync_with(din, a))
    elif isinstance(args, dict):
        for n, v in args.items():
            maybe_kwds[n] = sync_with(din, v)

    if maybe_args or maybe_kwds:
        dout = din | when_some | f(*maybe_args, **maybe_kwds)
    else:
        dout = din | when_some | f

    if balance is not None:
        ctrl = din['ctrl'] | balance
    else:
        ctrl = din['ctrl']

    return select(ctrl, Maybe[dout.dtype](), some(dout))


@gear
def sync_with(sync_in: Maybe, din, *, fcat=ccat, balance=None):
    return fcat(din, sync_in[1]) | Maybe | filt


@gear
def some(din) -> Maybe['din']:
    return ccat(din, Bool(True)) >> Maybe[din.dtype]


@gear
def when_some(din: Maybe) -> b'din.dtype':
    return din | filt(fixsel=1)



import inspect
import fnmatch
import importlib
import os
import re
import sys
import copy
from dataclasses import dataclass
from typing import Any, Callable

from .utils import (dict_generator, intercept_arguments, nested_get,
                    nested_set, safe_nested_set)

delimiter = '/'
wildcard_list = ['*', '?', '[', ']']


class Inject:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class MayInject:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


def get_args_from_registry(arg_dict):
    for k, v in arg_dict.items():
        if isinstance(v, Inject):
            arg_dict[k] = reg[v.args[0]]
        elif isinstance(v, MayInject):
            try:
                arg_dict[k] = reg[v.args[0]]
            except KeyError:
                arg_dict[k] = None


def inject_async(func):
    sig = inspect.signature(func)
    # default values in func definition
    injections = tuple(v.default.args[0] for k, v in sig.parameters.items()
                       if isinstance(v.default, Inject))

    func = intercept_arguments(func,
                               cb_named=get_args_from_registry,
                               cb_kwds=get_args_from_registry)

    try:
        all(reg[i] for i in injections)
        func()
    except KeyError:
        PluginBase.async_reg[func] = injections

    return func


def inject(func):
    return intercept_arguments(func,
                               cb_named=get_args_from_registry,
                               cb_kwds=get_args_from_registry)


class RegistryException(Exception):
    pass


class RegistryHook(dict):
    def __init__(self, **kwds):
        super().__init__()
        for k, v in kwds.items():
            self[k] = v

    def __getitem__(self, key):
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value):
        if not hasattr(self, f'_get_{key}'):
            dict.__setitem__(self, key, value)

        if hasattr(self, f'_set_{key}'):
            getattr(self, f'_set_{key}')(value)


class PluginBase:
    subclasses = []
    async_reg = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)
        cls.bind()

    @classmethod
    def clear(cls):
        pass

    @classmethod
    def purge(cls):
        cls.subclasses.clear()
        reg.clear()

    @classmethod
    def bind(cls):
        pass

    @classmethod
    def reset(cls):
        pass


def manage_async_regs(key_path):
    resolved = []

    for func, injections in PluginBase.async_reg.items():
        if key_path not in injections:
            continue

        try:
            all(reg[i] for i in injections)
            resolved.append(func)
            func()
        except KeyError:
            pass

    for r in resolved:
        PluginBase.async_reg.pop(r)


def reset():
    for subc in PluginBase.subclasses:
        subc.reset()


def clear():
    for subc in reversed(PluginBase.subclasses):
        subc.clear()

    reg.clear()

    for subc in PluginBase.subclasses:
        subc.bind()


@dataclass
class ConfigVariable:
    path: str
    default: Any
    docs: str = None
    setter: Callable = None
    getter: Callable = None

    def __post_init__(self):
        self._val = self.default

    @property
    def val(self):
        if self.getter is not None:
            return self.getter(self)

        return self._val

    @val.setter
    def val(self, val):
        if self.setter is not None:
            return self.setter(self, val)

        self._val = val

    @property
    def changed(self):
        return self._val != self.default


class Registry:
    def __init__(self, *args, **kwds):
        self._dict = dict(*args, **kwds)

    def __getitem__(self, key):
        key, _, subpath = key.partition('/')
        val = self._dict.__getitem__(key)

        if subpath:
            return val[subpath]

        if isinstance(val, ConfigVariable):
            return val.val

        return val

    def __iter__(self):
        return self._dict.__iter__()

    def items(self):
        return self._dict.items()

    def get(self, key, dflt):
        return self._dict.get(key, dflt)

    def clear(self):
        self._dict.clear()

    def keys(self):
        return self._dict.keys()

    def __contains__(self, key):
        key, _, subpath = key.partition('/')
        val = self._dict.__contains__(key)

        if subpath and val:
            return subpath in self._dict.__getitem__(key)

        return val

    def _setitem(self, key, val):
        key, _, subpath = key.partition('/')
        if not subpath:
            try:
                cfgvar = self._dict.__getitem__(key)
            except KeyError:
                pass
            else:
                if isinstance(cfgvar, ConfigVariable):
                    cfgvar.val = val
                    return

            return self._dict.__setitem__(key, val)

        if key not in self:
            subreg = Registry()
            self._dict.__setitem__(key, subreg)
        else:
            subreg = self._dict.__getitem__(key)

        subreg[subpath] = val

    def __setitem__(self, key, val):
        self._setitem(key, val)

        if self is reg:
            manage_async_regs(key)

    def confdef(self, path, default=None, docs=None, setter=None, getter=None):
        if path in self:
            raise Exception(f'Variable "{path}" already defined!')

        var = ConfigVariable(path,
                             default=default,
                             docs=docs,
                             setter=setter,
                             getter=getter)

        if setter is not None:
            setter(var, default)

        self[path] = var

        return var

    def subreg(self, key, val=None):
        if val is None:
            self[key] = Registry()
        else:
            self[key] = Registry(val)

        return self[key]


reg = Registry()


def load_plugin_folder(path, package=None):
    plugin_parent_dir, plugin_dir = os.path.split(path)
    sys.path.insert(0, plugin_parent_dir)
    pysearchre = re.compile('.py$', re.IGNORECASE)
    pluginfiles = filter(pysearchre.search, os.listdir(path))
    plugins = map(lambda fp: '.' + os.path.splitext(fp)[0], pluginfiles)
    # import parent module / namespace
    if package:
        importlib.import_module(plugin_dir, package=package)
    else:
        importlib.import_module(plugin_dir)

    modules = []
    for plugin in plugins:
        if not plugin.endswith('__'):
            if package:
                ret = importlib.import_module(
                    plugin, package=f'{package}.{plugin_dir}')
            else:
                ret = importlib.import_module(plugin, package=plugin_dir)

            modules.append(ret)

    return modules



from pygears import gear
from pygears.sim import clk
from pygears.util.utils import qrange


@gear(enablement='t.data == din')
async def parallelize(din, *, t) -> b't':
    data = t()

    for i, last in qrange(len(t)):
        async with din as val:
            data[i] = val

    await clk()
    yield data



from pygears import gear, alternative
from .fmaps import tuplemap
from .flatten import flatten
from .operators import code
from pygears.typing import Tuple, typeof, Queue, Union, Uint
from pygears.typing import factor as factor_type


def is_union_of_queues_with_equal_len(din):
    lengths = [t.width for t in din.types]
    return typeof(din, Union) and \
        all(typeof(t, Queue) for t in din.types) and \
        lengths[1:] == lengths[:-1]


# TODO: consider what to do with eots
@gear(enablement=b'is_union_of_queues_with_equal_len(din)')
def factor(din) -> b'factor(din)':
    lvl = din.dtype.types[0].lvl
    data_len = din.dtype.types[0].width - lvl
    din_tuple = din | Tuple \
        | tuplemap(f=(code(t=Tuple[Uint[data_len], Uint[lvl]]), None)) \
        | flatten

    return din_tuple[0, 2, 1] >> factor_type(din.dtype)


# TODO: consider what to do with eots
@alternative(factor)
@gear(enablement=b'typeof(din, Tuple) and all(typeof(t, Queue) for t in din)')
def factor_tuple_queue(din) -> b'factor(din)':
    din_tuple = din | tuplemap(f=(Tuple, )*len(din.dtype)) | flatten

    tuple_len = len(din_tuple.dtype)
    indices = list(range(0, tuple_len, 2))
    indices.append(1)  # eot

    return din_tuple[tuple(indices)] >> factor_type(din.dtype)



from pygears.conf import reg
from pygears.typing.base import Any, GenericMeta, type_repr, typeof, T


class TypeMatchError(Exception):
    pass


def _get_match_conds_rec(t, pat, matches):
    # Ignore type template arguments, i.e.: 'T2' in Tuple[1, 2, 3, 'T2']
    if isinstance(t, str):
        return t

    if (t == Any) or (pat == Any):
        return t

    if t == pat:
        return t

    if isinstance(pat, T):
        t = _get_match_conds_rec(t, pat.__bound__, matches)
        pat = pat.__name__

    # Did we reach the parameter name?
    if isinstance(pat, str):
        if pat in matches:
            # If the parameter name is already bound, check if two deductions
            # are same
            # if repr(t) != repr(matches[pat]) and t != Any and matches[pat] != Any:
            if t != matches[pat] and t != Any and matches[pat] != Any:
                raise TypeMatchError(
                    f'Ambiguous match for parameter "{pat}": {type_repr(t)} '
                    f"and {type_repr(matches[pat])}")
        else:
            try:
                # TODO: Should probably use globals of the string. See: Python
                # 3.10. typing.get_type_hints()
                res = eval(pat, reg['gear/type_arith'], matches)
                if t != res:
                    raise TypeMatchError(
                        f"{type_repr(t)} cannot be matched to {type_repr(res)}")
            except Exception as e:
                matches[pat] = t

        return t

    if not (isinstance(t, GenericMeta) and isinstance(pat, GenericMeta)
            and typeof(t.base, pat.base)):
        raise TypeMatchError(
            "{} cannot be matched to {}".format(type_repr(t), type_repr(pat)))

    # TODO: There might be multiple levels of inheritance when the types are
    # created, maybe they are compatible base on some of their more distant
    # base classes
    if pat.args:
        if len(t.args) != len(pat.args):
            raise TypeMatchError(
                "{} cannot be matched to {}".format(type_repr(t), type_repr(pat)))

        args = []
        for ta, pa in zip(t.args, pat.args):
            try:
                res = _get_match_conds_rec(ta, pa, matches)
                args.append(res)
            except TypeMatchError as e:
                raise TypeMatchError(
                    f'{str(e)}\n - when matching {repr(t)} to {repr(pat)}')

        # if hasattr(pat, '__parameters__'):
        args = {name: a for name, a in zip(pat.fields, args)}
    else:
        args = t.args

    # TODO: Revisit this Don't create a new type class when class has no
    # specified templates, so that we don't end up with multiple different
    # base class objects, that cannot be correctly tested with "issubclass"
    if not args and not t.args:
        return t
    else:
        return t.__class__(t.__name__, t.__bases__, dict(t.__dict__), args=args)


def get_match_conds(t, pat, matches=None):
    if matches is None:
        matches = {}
    else:
        matches = dict(matches)

    res = _get_match_conds_rec(t, pat, matches)
    return matches, res


def match(t, pat, matches=None):
    try:
        upd_matches, res = get_match_conds(t, pat, matches)
    except TypeMatchError:
        return None

    if matches is not None:
        matches.clear()
        matches.update(upd_matches)

    return res



# from pygears.hls.hls_expressions import EXTENDABLE_OPERATORS, BinOpExpr
from pygears.hdl.sv.sv_expression import SVExpressionVisitor
from pygears.typing import Array, Number, typeof, is_type

RESIZE_FUNC_TEMPLATE = """
function {signed} [{res_size}:0] {name};
    input {signed} [{op_size}:0] tmp;
    begin
        {name} = tmp;
    end
endfunction
"""


def get_resize_func(res_dtype, op_dtype):
    res_size = res_dtype.width
    op_size = op_dtype.width
    signed = 'signed' if getattr(op_dtype, 'signed', False) else ''

    name = f'resize_{op_size}_to_{res_size}_{signed}'
    val = RESIZE_FUNC_TEMPLATE.format(res_size=res_size-1,
                                      op_size=op_size-1,
                                      signed=signed,
                                      name=name)
    return name, val


def update_extras(all_extras, addition):
    if addition:
        for key, value in addition.items():
            if key:
                all_extras[key] = value


def cast(res_dtype, op_dtype, op_value):
    truncate_func = None
    truncate_impl = None

    if isinstance(op_value, int):
        return str(op_value), {truncate_func: truncate_impl}

    res = op_value

    if op_dtype.width != res_dtype.width:
        truncate_func, truncate_impl = get_resize_func(res_dtype, op_dtype)
        res = f'{truncate_func}({res})'

    res_signed = getattr(res_dtype, 'signed', False)
    op_signed = getattr(op_dtype, 'signed', False)

    if res_signed != op_signed:
        sign = '$signed' if res_signed else '$unsigned'
        res = f'{sign}({res})'

    return res, {truncate_func: truncate_impl}


class VExpressionVisitor(SVExpressionVisitor):
    def __init__(self):
        super(VExpressionVisitor, self).__init__()
        self.merge_with = '_'
        self.expr = vexpr
        self.extras = {}

    def visit_CastExpr(self, node):
        op = self.visit(node.operand)
        val, func = cast(node.cast_to, node.operand.dtype, op)
        update_extras(self.extras, func)
        return val

    def visit_BinOpExpr(self, node):
        ops = [self.visit(op) for op in node.operands]
        for i, op in enumerate(node.operands):
            if isinstance(op, BinOpExpr):
                ops[i] = f'({ops[i]})'

        if node.operator not in EXTENDABLE_OPERATORS:
            return f'{ops[0]} {node.operator} {ops[1]}'

        res_dtype = node.dtype
        if node.operands[0].dtype.width > res_dtype.width:
            res_dtype = node.operands[0].dtype
        if node.operands[1].dtype.width > res_dtype.width:
            res_dtype = node.operands[1].dtype

        val0, func = cast(res_dtype, node.operands[0].dtype, ops[0])
        update_extras(self.extras, func)
        val1, func = cast(res_dtype, node.operands[1].dtype, ops[1])
        update_extras(self.extras, func)

        return val0 + f" {node.operator} " + val1

    def visit_SubscriptExpr(self, node):
        val = self.visit(node.val)

        if isinstance(node.index, slice):
            return f'{val}[{int(node.index.stop) - 1}:{node.index.start}]'

        dtype = node.val.dtype
        if typeof(dtype, Array):
            index = self.visit(node.index)
            return f'{val}_arr[{index}]'
        elif typeof(dtype, Number):
            return f'{val}[{self.visit(node.index)}]'
        elif is_type(dtype):
            return f'{val}_{dtype.fields[node.index]}'
        else:
            raise Exception('Unable to subscript')


def vexpr(expr, extras=None):
    v_visit = VExpressionVisitor()
    res = v_visit.visit(expr)
    if extras is not None:
        update_extras(extras, v_visit.extras)
    return res



from pygears import gear
from pygears.util.utils import gather
from pygears.lib.shred import shred
from pygears.typing import Tuple


# @gear
@gear
async def ccat(*din) -> b'Tuple[din]':
    """Short for concatenate, combines multiple interfaces into a single interface
    whose type is a :class:`Tuple` of the input interface types. One output
    data is formed by combining one data from each of the inputs::

        din0 = Intf(Uint[8])
        din1 = Intf(Uint[16])

    >>> ccat(din0, din1)
    Intf(Tuple[Uint[8], Uint[16]])

    """

    yield [await d.pull() for d in din]

    for d in din:
        d.ack()

    # async with gather(*din) as dout:
    #     yield dout


@gear
def ccat_sync_with(sync_in, din, *, balance=None):
    if balance:
        sync_in = sync_in | balance

    din_sync, sync_in_sync = ccat(din, sync_in)
    sync_in_sync | shred

    return din_sync



from pygears import reg
from pygears.sim import log
from pygears.sim.extens.sim_extend import SimExtend
from pygears.typing import Queue
from pygears.conf import inject, Inject


class RandBase(SimExtend):
    @inject
    def __init__(self, cons, outdir=Inject('results-dir')):
        super().__init__()
        self.outdir = outdir
        self.constraints = self.randomizes(cons)
        reg['sim/config/randomizer'] = self

    def get_dtype_by_name(self, name):
        for constraint in self.constraints:
            if constraint.name == name:
                return constraint.dtype

        # check if queue
        data = None
        eot = None
        for constraint in self.constraints:
            if constraint.name == f'{name}_data':
                data = constraint.dtype
            if constraint.name == f'{name}_eot':
                eot = constraint.dtype

        if data and eot:
            return Queue[data, len(eot)]

    def randomizes(self, cons):
        constraints = []
        for c in cons:
            constraints.append(self.create_type_cons(c.get_data_desc()))
            if c.is_queue:
                constraints.append(self.create_type_cons(c.get_eot_desc()))
        return constraints

    def create_type_cons(self, desc={}):
        log.error('Create type constraints function not implemented.')

    def get_rand(self, name):
        log.error('Get rand function not implemented.')



import functools
import inspect
import sys
from pygears.conf import MultiAlternativeError
from pygears.typing.base import hashabledict


def extract_arg_kwds(kwds, func):
    try:
        arg_names, _, varkw, _, kwonlyargs, *_ = inspect.getfullargspec(func)
    except TypeError:
        # Function not inspectable (probably a builtin), do nothing
        return {}, kwds

    arg_kwds = {}
    kwds_only = {}
    for k in list(kwds.keys()):
        if k in arg_names:
            arg_kwds[k] = kwds[k]
        elif (k in kwonlyargs) or (varkw is not None):
            kwds_only[k] = kwds[k]
        else:
            kwds_only[k] = kwds[k]
            raise TypeError(f"{func.__name__}() got an unexpected keyword argument '{k}'")

    return arg_kwds, kwds_only


def combine_arg_kwds(args, kwds, func):
    try:
        arg_names, varargs, *_ = inspect.getfullargspec(func)
    except TypeError:
        # Function not inspectable (probably a builtin), do nothing
        return args

    if varargs:
        return args

    args_unmatched = list(args)
    args_comb = []
    for a in arg_names:
        if a in kwds:
            args_comb.append(kwds[a])
        elif args_unmatched:
            args_comb.append(args_unmatched.pop(0))
        else:
            break

    # If some args could not be matched to argument names, raise an Exception
    if args_unmatched:
        raise TypeError(f"Too many positional arguments for {func.__name__}()")

    return args_comb


def all_args_specified(args, func):
    arg_names, varargs, _, _, kwds, _, types = inspect.getfullargspec(func)
    if varargs:
        return len(args) > 0
    elif len(args) == len(arg_names):
        return True
    else:
        return False


class Partial:
    '''The Partial class implements a mechanism similar to that of the
    functools.partial, with one important difference with how calling
    its objects operates.

    When Partial class is instantiated, arguments that are to be passed to the
    wrapped function are inspected. If not enough positional arguments, as
    demanded by the wrapped function, have been supplied, a Partial object is
    returned that remembers which arguments have been supplied. Otherwise, if
    enough arguments have been supplied, instead of returning a new Partial
    object, the wrapped function is called with the supplied arguments and its
    return value is returned instead.

    When instantiated Partial object is called (via __call__ method), the
    arguments supplied to the call are combined with the arguments already
    supplied via Partial object constructor and Partial class is then called
    with a combined set of arguments to either return a new Partial object or
    call the wrapped function as described by the above paragraph

    This class also supports supplying arguments by pipe '|' operator.

    '''
    def __init__(self, func, *args, **kwds):
        functools.update_wrapper(self, func)
        self.func = func
        self.args = args
        self.kwds = kwds
        self.errors = []
        self._cache = {}

    def __str__(self):
        return f'{self.func.__name__}'

    def __str__(self):
        return f'{self.func.__name__}'

    def __call__(self, *args, **kwds):
        if self.kwds:
            prev_kwds = self.kwds.copy()
            prev_kwds.update(kwds)
            kwds = prev_kwds

        if self.args:
            args = self.args + args

        try:
            key = hash(functools._make_key(args, kwds, False)) ^ hash(self.func)
        except TypeError:
            key = None

        if key in self._cache:
            return self._cache[key](*args, **kwds)

        no_unpack_alt = kwds.pop('__no_unpack_alt__', False)
        alternatives = [self.func] + getattr(self.func, 'alternatives', [])
        if no_unpack_alt:
            alternatives = [f for f in alternatives if not f.__name__.endswith('_unpack__')]

        errors = [None] * len(alternatives)

        for i, func in enumerate(alternatives):
            try:
                kwd_intfs, kwd_params = extract_arg_kwds(kwds, func)
                args_comb = combine_arg_kwds(args, kwd_intfs, func)

                if all_args_specified(args_comb, func):
                    ret = func(*args_comb, **kwd_params)

                    if key is not None:
                        self._cache[key] = func

                    errors = []
                    return ret
                else:
                    # TODO: Can happen if user forgets '*' for separation, warn about this
                    # TODO: Think about disabling keyword arguments that are not keyword-only for gears
                    msg = f"not enough arguments specified for '{self.func.__name__}'"
                    if hasattr(func, 'alternative_to'):
                        arg_signature = ', '.join(
                            f'{name}: {repr(dtype)}'
                            for name, dtype in func.__annotations__.items())

                        alt_arg_names, *_ = inspect.getfullargspec(func.alternative_to)

                        msg = (
                            f"not enough arguments specified for '{self.func.__name__}' for an"
                            f" alternative with argument '{alt_arg_names[0]}' unpacked to: ({arg_signature})"
                        )

                    errors[i] = (func, TypeError, TypeError(msg), [])
            except Exception as e:
                # If no alternatives, just re-raise an error
                if len(alternatives) == 1:
                    raise e
                else:
                    # errors.append((func, e, sys.exc_info()))

                    errors[i] = (func, *sys.exc_info())
        else:
            if self._all_alternative_error(errors):
                raise MultiAlternativeError(errors)
            else:
                # If some alternative can handle more arguments, try to wait
                # for it
                p = Partial(self.func, *args, **kwds)
                p.errors = errors[:]
                return p

    def _all_alternative_error(self, errors):
        if errors.count(None) != 0:
            return False

        for e in errors:
            if e[-1] == []:
                return False

        return True

    def __or__(self, iin):
        return iin.__ror__(self)

    def __ror__(self, iin):
        if type(iin) == tuple:
            return self(*iin)
        else:
            return self(iin)



from pygears.hls.ir import SubscriptExpr
from pygears import reg
from pygears.sim import clk
from ..ir_utils import Scope, HDLVisitor, res_true, add_to_list, ir, res_false, IrExprRewriter, IrExprVisitor
from .. import HLSSyntaxError
from pygears.typing import cast
from ..cfg import CfgDfs
from ..cfg_util import remove_node
from .exit_cond_cfg import cond_wrap
from copy import copy


def get_forward_value(target, scope):
    if isinstance(target, ir.Name):
        if target.name not in scope:
            if target.obj.reg:
                return None

        return scope[target.name]
    elif isinstance(target, ir.SubscriptExpr):
        if isinstance(target.index, ir.ResExpr):
            base_val = get_forward_value(target.val)
            if base_val is None:
                return None

            return base_val[target.index.val]
        else:
            return None


def forward_value(target, val, scope):
    if isinstance(target, ir.Name):
        scope[target.name] = val
        return True
    elif isinstance(target, ir.Component):
        scope[f'{target.val.name}.{target.field}'] = val
        return True
    elif isinstance(target, ir.ConcatExpr):
        for i, t in enumerate(target.operands):
            forward_value(t, ir.SubscriptExpr(val, ir.ResExpr(i)), scope)
    elif isinstance(target, ir.SubscriptExpr):
        if (isinstance(target.index, ir.ResExpr) and isinstance(val, ir.ResExpr)):
            base_val = get_forward_value(target.val, scope)

            if isinstance(base_val, ir.ResExpr):
                base_val.val[target.index.val] = cast(val.val, base_val.dtype[target.index.val])
                return True
        else:
            scope[target.val.name] = target.val


def merge_subscope(subscopes, tests):
    names = set()
    for s in subscopes:
        names |= set(s.keys())

    outscope = {}
    for n in names:
        vals = []
        for s in subscopes:
            if n.endswith(']') and n not in s:
                vect_val = s.get(n[:n.index('[')], None)
                if vect_val is not None:
                    index = int(n[n.index('[')+1:-1])
                    vals.append(ir.SubscriptExpr(vect_val, ir.ResExpr(index)))
                else:
                    vals.append(None)
            else:
                vals.append(s.get(n, None))

        if vals.count(vals[0]) == len(vals):
            outscope[n] = vals[0]
            continue

        prev_val = None
        for v, t in zip(reversed(vals), reversed(tests)):
            if prev_val is None:
                # if b.value.test != ir.res_true and not n.endswith('.data'):
                #     raise HLSSyntaxError(f'Variable "{n}" uninitialized in some cases')

                # TODO: Assigning to output interface is more like an event and
                # less like assigning to a variable, so the following check is not valid
                # There should be a better way of handling outputs
                # TODO: Think when iit is OK to have variable initialized only in one branch?
                if t != ir.res_true:
                    prev_val = None
                else:
                    prev_val = v
            elif v is None:
                # TODO: Connected with two todos above, result of possibly uninitialized variable
                prev_val = None
            else:
                prev_val = ir.ConditionalExpr((v, prev_val), t)

        # TODO: Connected with two todos above, result of possibly uninitialized variable
        if prev_val is not None:
            outscope[n] = prev_val

    return outscope


def cfg_merge_subscope(block):
    subscopes = [b.scope for b in block.next] + [block.scope]
    tests = [b.test for b in block.next] + [ir.res_true]

    return merge_subscope(subscopes, tests)


class DependencyVisitor(IrExprVisitor):
    def __init__(self):
        self.variables = set()

    def visit_Name(self, irnode):
        self.variables.add(irnode.name)


def expr_dependencies(node):
    v = DependencyVisitor()
    v.visit(node)
    return v.variables


class Inliner(IrExprRewriter):
    def __init__(self, scope, ctx, missing_ok=False, reg_inits=None):
        self.scope_map = scope
        self.ctx = ctx
        self.missing_ok = missing_ok

        if reg_inits is None:
            reg_inits = {}
        self.reg_inits = reg_inits

    def visit_SubscriptExpr(self, node):
        if node.ctx == 'load':
            return super().visit_SubscriptExpr(node)
        else:
            return ir.SubscriptExpr(node.val, self.visit(node.index), ctx=node.ctx)

    def visit_Name(self, irnode):
        # If this name is target of the assignment, we have nothing to do
        if (irnode.ctx != 'load'):
            return None

        if (irnode.name not in self.scope_map):
            if self.missing_ok:
                return irnode

        val = self.scope_map[irnode.name]

        if any(k.startswith(f'{irnode.name}[') for k in self.scope_map):
            elems = [ir.SubscriptExpr(irnode, ir.ResExpr(i)) for i in range(len(irnode.dtype))]
            for n in self.scope_map:
                if not n.startswith(f'{irnode.name}['):
                    continue

                index = int(n[n.index('[')+1:-1])
                elems[index] = self.scope_map[n]

            val = ir.ConcatExpr(elems)

        if isinstance(val, ir.Name) and val.obj.reg and val.name in self.reg_inits:
            return val.obj.val

        # TODO: What's with unknown?
        if isinstance(val, ir.ResExpr) and getattr(val.val, 'unknown', False):
            return irnode

        return val


def inline_expr(irnode, scope, ctx, reg_inits=None):
    new_node = Inliner(scope, ctx, reg_inits=reg_inits).visit(irnode)
    if new_node is None:
        return irnode

    return new_node


def detect_new_state(node, scope_map):
    # TODO: Maybe break state at better boundaries, like the start of a loop if
    # the loop is blocking and we know we need a new state for that
    expr = node.value.expr
    forward = scope_map['forward']

    if expr == 'forward':
        return ir.UnaryOpExpr(forward, ir.opc.Not)
    elif expr == 'back':
        scope_map['forward'] = ir.res_false
        return ir.res_false
    elif expr == ir.ResExpr(clk):
        node.value.expr = 'forward'
        return ir.res_true
    elif expr == 'break':
        return ir.res_false

    if expr.field == 'valid':
        return ir.UnaryOpExpr(forward, ir.opc.Not)

    if expr.field == 'ready':
        scope_map['forward'] = ir.res_false

    return ir.res_false


class Scoping(CfgDfs):
    def __init__(self, ctx, scope_map=None):
        super().__init__()
        self.ctx = ctx
        self.scope_map = {} if scope_map is None else scope_map

    def handle_expr(self, expr):
        return inline_expr(expr, self.scope_map, self.ctx)

    def enter_FuncBlock(self, block):
        for a in block.value.args:
            self.scope_map[a] = self.ctx.ref(a)

    def enter_Statement(self, block):
        block.scope = copy(self.scope_map)

    def exit_HDLBlock(self, block):
        outscope = cfg_merge_subscope(block)
        self.scope_map.update(outscope)

    def enter_Branch(self, node):
        self.scope_map = copy(node.prev[0].scope)
        irnode: ir.Branch = node.value
        node.test = self.handle_expr(irnode.test)

    def exit_Branch(self, node):
        node.scope = copy(self.scope_map)

    def enter_AssignValue(self, node):
        irnode: ir.AssignValue = node.value

        val = self.handle_expr(irnode.val)

        if isinstance(val, ir.ConcatExpr):
            val = ir.ConcatExpr(
                operands=[op.expr if isinstance(op, ir.Await) else op for op in val.operands])

        forward_value(irnode.target, val, self.scope_map)
        node.val = val


class Inline(Scoping):
    def handle_expr(self, expr):
        return inline_expr(expr, self.scope_map, self.ctx)

    def enter_FuncReturn(self, node):
        irnode: ir.FuncReturn = node.value
        irnode.expr = self.handle_expr(irnode.expr)

    def enter_Branch(self, node):
        super().enter_Branch(node)
        irnode: ir.Branch = node.value
        irnode.test = node.test

    def enter_AssignValue(self, node):
        super().enter_AssignValue(node)
        irnode: ir.AssignValue = node.value
        irnode.val = node.val


class VarScope(Inline):
    def __init__(self, ctx, state_in_scope=None, state_id=None, new_states=None):
        if state_in_scope:
            self.scope_map = {}
            self.scope_map.update(state_in_scope[state_id])
            for i in ctx.intfs:
                self.scope_map[i] = ctx.ref(i)
                if f'{i}.valid' not in self.scope_map:
                    self.scope_map[f'{i}.valid'] = ir.res_false

            for r in ctx.regs:
                self.scope_map[r] = ctx.ref(r)

            self.scope_map['_state'] = ctx.ref('_state')
            self.scope_map['forward'] = ir.res_true
        else:
            self.scope_map = {}

        self.new_states = new_states
        self.state_in_scope = state_in_scope
        self.state_id = state_id
        self.reg_inits = set()
        super().__init__(ctx, self.scope_map)

    def within_loop(self):
        for s in reversed(self.scopes):
            if isinstance(s.value, ir.LoopBody):
                return True
        else:
            return False

    def handle_expr(self, expr):
        # TODO: Think of more general heuristic, when it is better to
        # substitute a register with its initial value. This should probably
        # omitted only within the loop that updates the register value?
        if self.within_loop():
            return inline_expr(expr, self.scope_map, self.ctx)
        else:
            return inline_expr(expr, self.scope_map, self.ctx, self.reg_inits)

    def transition_scope(self, exit_node, state_id):
        if state_id == self.state_id:
            return

        in_scope = {}
        in_scope = copy(self.ctx.intfs)
        in_scope['_state'] = self.ctx.scope['_state']

        for name, _ in self.ctx.reaching[id(exit_node.value)]['out']:
            if name in self.ctx.regs:
                in_scope[name] = self.ctx.ref(name)
            # TODO: check this, it fails on 'qrange_dout.ready' for an example
            elif name in self.scope_map:
                in_scope[name] = self.scope_map[name]

        self.state_in_scope[state_id] = in_scope

    def enter_Jump(self, node):
        irnode: ir.Jump = node.value
        if irnode.label == 'state':
            state_id = irnode.where
            self.transition_scope(node, state_id)

    def enter_RegReset(self, node):
        self.ctx.reset_states[node.value.target.name].add(self.state_id)
        self.reg_inits.add(node.value.target.name)
        # remove_node(node)

    def enter_Await(self, node):
        irnode: ir.Await = node.value

        if irnode.expr == ir.res_false:
            return

        if isinstance(irnode.expr, ir.Component) and irnode.expr.field == 'ready':
            self.scope_map[f'{irnode.expr.val.name}.ready'] = ir.ResExpr(True)

        cond = detect_new_state(node, self.scope_map)
        if cond != ir.res_false:
            # print(f'New state cond: {str(cond)}')
            state_id = len(self.state_in_scope)
            self.state_in_scope.append(None)

            self.transition_scope(node, state_id)
            self.new_states[node] = cond

        if cond == ir.res_true:
            return True


class ExploreState(Inline):
    def __init__(self, ctx, scope_map, state_id):
        self.states = []
        self.state_id = state_id
        self.reg_inits = set()

        scope_map = scope_map.copy()
        for i in ctx.intfs:
            scope_map[i] = ctx.ref(i)
            if f'{i}.valid' not in scope_map:
                scope_map[f'{i}.valid'] = ir.res_false

        for r in ctx.regs:
            scope_map[r] = ctx.ref(r)

        scope_map['_state'] = ctx.ref('_state')
        scope_map['forward'] = ir.res_true

        super().__init__(ctx, scope_map)

    def within_loop(self):
        for s in reversed(self.scopes):
            if isinstance(s.value, ir.LoopBody):
                return True
        else:
            return False

    def handle_expr(self, expr):
        # TODO: Think of more general heuristic, when it is better to
        # substitute a register with its initial value. This should probably
        # omitted only within the loop that updates the register value?
        if self.within_loop():
            return inline_expr(expr, self.scope_map, self.ctx)
        else:
            return inline_expr(expr, self.scope_map, self.ctx, self.reg_inits)

    def enter_RegReset(self, node):
        self.reg_inits.add(node.value.target.name)

    def enter_Jump(self, node):
        irnode: ir.Jump = node.value
        if irnode.label == 'state':
            state_id = irnode.where
            for s in reversed(self.scopes):
                if (isinstance(s.value, ir.LoopBody) and s.value.state_id == state_id
                        and state_id != self.state_id):
                    self.states.append(state_id)

    def enter_Await(self, node):
        irnode: ir.Await = node.value

        if irnode.expr == ir.res_false:
            return

        if isinstance(irnode.expr, ir.Component) and irnode.expr.field == 'ready':
            self.scope_map[f'{irnode.expr.val.name}.ready'] = ir.ResExpr(True)

        cond = detect_new_state(node, self.scope_map)

        if cond == ir.res_true:
            return True



import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

LIB_SVLIB_DIR = os.path.join(ROOT_DIR, 'lib', 'svlib')
USER_SVLIB_DIR = os.path.expanduser('~/.pygears/svlib')

LIB_VLIB_DIR = os.path.join(ROOT_DIR, 'lib', 'vlib')
USER_VLIB_DIR = os.path.expanduser('~/.pygears/vlib')

CACHE_DIR = os.path.join(
    os.environ.get('XDG_CACHE_HOME',
                   os.path.join(os.path.expandvars("$HOME"), '.cache')),
    'pygears')



from enum import IntEnum
import time
import multiprocessing
from pygears.core.hier_node import HierYielderBase
from pygears.conf import inject, Inject
import os
from pygears.typing import typeof
from pygears import find, reg
from pygears.hdl import HDLPlugin
from pygears.core.gear import Gear, Intf
from pygears.core.port import HDLConsumer, Port, InPort
from pygears.sim import timestep
from .sim_extend import SimExtend
from pygears.core.hier_node import HierVisitorBase
from pygears.sim.modules import SimVerilated, SimSocket

multiprocessing.set_start_method('fork')

VALUE = set(('0', '1', 'x', 'X', 'z', 'Z'))
INVALID_VALUE = set(('x', 'X', 'z', 'Z'))
VECTOR_VALUE_CHANGE = set(('b', 'B', 'r', 'R'))
VALUE_CHANGE = VALUE | VECTOR_VALUE_CHANGE


@inject
def find_cosim_modules(top=Inject('gear/root')):
    class CosimVisitor(HierVisitorBase):
        @inject
        def __init__(self, sim_map=Inject('sim/map')):
            self.sim_map = sim_map
            self.cosim_modules = []

        def Gear(self, module):
            if isinstance(self.sim_map.get(module, None), SimVerilated):
                self.cosim_modules.append(self.sim_map[module])
                return True

    v = CosimVisitor()
    v.visit(top)
    return v.cosim_modules


class ChannelState(IntEnum):
    Invalid = 0
    NotReady = 1
    Ready = 2
    Done = 3
    Awaiting = 4


import functools


@functools.lru_cache(maxsize=None)
def subtypes(dtype):
    return tuple((t, t.width) for t in dtype)


def split_coded_dtype(t, val):
    for subt, subt_width in t:
        subt_mask = (1 << subt_width) - 1
        yield val & subt_mask
        val >>= subt_width


def split_coded_change(t, val1, val2):
    for subt, subt_width in subtypes(t):
        subt_mask = (1 << subt_width) - 1

        subval1 = val1 & subt_mask
        val1 >>= subt_width

        if val2 is not None:
            subval2 = val2 & subt_mask
            val2 >>= subt_width
        else:
            subval2 = None

        yield subt, subval1, subval2


def create_data_change(t, val, prev_val):
    from pygears.typing import Queue, Array, Tuple
    is_changed = prev_val is None and val is not None or val != prev_val

    if typeof(t, (Queue, Array, Tuple)):
        change = [
            create_data_change(subt, v, prev_v)
            for subt, v, prev_v in split_coded_change(t, val, prev_val)
        ]

        # return {'isValueComplex': True, 'isDataChanged': is_changed, 'value': change}
        return (1, int(is_changed), change)
    else:
        if isinstance(val, (int, float)):
            val = int(val)

        return (0, int(is_changed), val)
        # return {'isValueComplex': False, 'isDataChanged': is_changed, 'value': val}


class VcdToJson:
    def __init__(self):
        self.json_vcd = {}
        self.diff = {}
        self.state = {}
        self.value = {}

    def create_change(self, timestep, state, state_change, t, val, prev_val):
        # elem = {
        #     'cycle': timestep,
        #     'state': int(state),
        #     'isStateChanged': state_change,
        # }

        elem = [timestep, int(state), int(state_change)]

        if val is not None:
            elem.append(create_data_change(t, val, prev_val))
            # elem['data'] = create_data_change(t, val, prev_val)

        import json
        # print(json.dumps(elem, separators=(',', ':')))

        return json.dumps(elem, separators=(',', ':'))

    def after_timestep(self, timestep):
        for p, d in self.diff.items():
            changes = self.json_vcd[p.name]
            data_change = False
            new_state = state = self.state[p]
            prev_val = self.value[p]
            new_val = d['d']

            # data_change = new_val is not None and (prev_val or prev_val != new_val)

            # if data_change:
            #     new_val = p.dtype.decode(d['d'])
            # else:
            #     new_val = None

            if state == ChannelState.Invalid:
                if d['v'] and d['r']:
                    new_state = ChannelState.Ready
                    data_change = True
                elif d['v']:
                    new_state = ChannelState.NotReady
                    data_change = True
                elif d['r']:
                    new_state = ChannelState.Awaiting
            elif state == ChannelState.Ready:
                if d['r'] == 0 and d['v'] == 0:
                    new_state = ChannelState.Invalid
                elif d['r'] == 0:
                    new_state = ChannelState.NotReady
                    data_change = prev_val != new_val
                elif d['v'] == 0:
                    new_state = ChannelState.Awaiting
                else:
                    data_change = prev_val != new_val
            elif state == ChannelState.NotReady:
                if d['r']:
                    new_state = ChannelState.Ready
            elif state == ChannelState.Awaiting:
                if d['r'] == 0 and d['v']:
                    new_state = ChannelState.NotReady
                    data_change = prev_val != new_val
                elif d['v']:
                    new_state = ChannelState.Ready
                    data_change = prev_val != new_val

            if new_state != state or data_change or timestep == 0:
                cycle_change = self.create_change(
                    timestep,
                    new_state,
                    state_change=(new_state != state),
                    t=p.dtype,
                    val=(None if new_state == ChannelState.Invalid else new_val),
                    prev_val=prev_val)

                if cycle_change is not None:
                    changes.append(cycle_change)

            if data_change:
                self.value[p] = new_val

            self.state[p] = new_state

        # self.diff.clear()


def follow(fn, finish_event, sleep_sec=0.1):
    """ Yield each line from a file as they are written.
    `sleep_sec` is the time to sleep after empty reads. """
    line = ''
    with open(fn) as f:
        while True:
            tmp = f.readline()
            if tmp:
                line += tmp
                if line.endswith("\n"):
                    yield line
                    line = ''
            else:
                if finish_event.is_set():
                    return

                time.sleep(sleep_sec)


# def follow(fn, finish_event, sleep_sec=0.5):
#     import time
#     import subprocess
#     import select

#     f = subprocess.Popen(['tail', '-F', fn], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#     p = select.poll()
#     p.register(f.stdout)

#     while True:
#         if p.poll(sleep_sec):
#             yield f.stdout.readline().decode()
#         elif finish_event.is_set():
#             return
#         else:
#             # print(f'Sleep')
#             time.sleep(sleep_sec)


def vcd_to_json_worker(entries, wire_map: dict, vcd_conv, t):
    for identifier_code, value in entries:

        if identifier_code not in wire_map:
            continue

        for port, wire_name in wire_map[identifier_code]:
            # print(f'"{identifier_code}": {port.name}_{wire_name} change @ {t}')

            if port not in vcd_conv.diff:
                vcd_conv.diff[port] = {'r': None, 'v': None, 'd': None}

            if wire_name == 'valid':
                vcd_conv.diff[port]['v'] = value
            elif wire_name == 'ready':
                vcd_conv.diff[port]['r'] = value
            elif wire_name == 'data' and value is not None:
                vcd_conv.diff[port]['d'] = value

    vcd_conv.after_timestep(t)


def vcd_to_json(vcd_fn, finish_event, ret_pipe, top):
    import os
    import time

    while not os.path.exists(vcd_fn):
        time.sleep(0.1)

    time.sleep(0.1)

    wire_map = {}
    vcd_conv = None
    skip_scope = 0

    t = -1
    hier = top
    worker_data = []

    for line in follow(vcd_fn, finish_event):
        # for line in f:
        line = line.strip()
        if line == '':
            continue

        line_head = line[0]

        if line_head == '#':
            next_t = int(line[1:])
            if next_t % 10 != 0:
                continue

            next_t //= 10

            if t == -1:
                vcd_conv = VcdToJson()
                for identifier_code in wire_map:
                    for p, _ in wire_map[identifier_code]:
                        vcd_conv.state[p] = ChannelState.Invalid
                        vcd_conv.value[p] = None
                        vcd_conv.json_vcd[p.name] = []

            if t >= 0:
                vcd_to_json_worker(worker_data, wire_map, vcd_conv, t)
                worker_data.clear()

            t = next_t

        elif line_head in VALUE_CHANGE:
            if line_head in VECTOR_VALUE_CHANGE:
                value, identifier_code = line[1:].split()
            elif line_head in VALUE:
                value = line[0]
                identifier_code = line[1:]

            if value[0] in INVALID_VALUE:
                value = None
            elif value[0] not in ('b', 'B'):
                value = int(value, 2)
            else:
                value = int(value[1:])

            if identifier_code in wire_map:
                worker_data.append((identifier_code, value))

        elif '$enddefinitions' in line:
            pass
        elif '$scope' in line:
            if skip_scope:
                skip_scope += 1
                continue

            segs = line.split()
            # Name of the TOP is ' '
            if len(segs) == 3:
                hier = top
                continue

            scope_name = segs[2]
            child = None

            if scope_name.startswith('_') and scope_name.endswith('_spy'):
                scope_name = scope_name[1:-4]
                child = find(f'{hier.name}.{scope_name}')

            if scope_name.startswith('bc_') or '_bc_' in scope_name:
                skip_scope = 1
                continue

            if scope_name == 'TOP' or scope_name.endswith('_wrap'):
                hier = top
                continue

            if child is None:
                child = find(f'{hier.name}.{scope_name}')
                if not isinstance(child, Port):
                    child = None

            if child is None:
                child = find(f'{hier.name}/{scope_name}')

            if child is None:
                skip_scope = 1
                continue

            # if child is None:
            #     if child is None:
            #         skip_scope = 1
            #         continue

            #     child = hier[scope_name]
            #     # if child is not None and not child.hierarchical:
            #     #     skip_scope = 1
            #     #     continue

            #     # print(f'Gear scope: {f"{hier.name}/{scope_name}"}')
            # # else:
            # #     print(f'Port scope: {f"{hier.name}.{scope_name}"}')

            # if child is None:
            #     raise Exception(f'Cannot find gear for scope: {line}')

            hier = child

        elif '$upscope' in line:
            if skip_scope:
                skip_scope -= 1
                continue

            if isinstance(hier, (Gear, Intf)):
                if hier is not top:
                    hier = hier.parent
            else:
                hier = hier.gear
        elif '$var' in line:
            # print(line)
            if hier is top or skip_scope:
                continue

            if not isinstance(hier, (Port, Intf)):
                continue

            if isinstance(hier,
                          Port) and not isinstance(hier, InPort) and not hier.gear.hierarchical:
                continue

            ls = line.split()
            identifier_code = ls[3]
            name = ls[4]
            # name = ''.join(ls[4:-1])

            # print(f'Var "{identifier_code}": {hier.name}_{name}')
            if identifier_code not in wire_map:
                wire_map[identifier_code] = []

            wire_map[identifier_code].append((hier, name))

        elif '$timescale' in line:
            continue

    if vcd_conv is None:
        raise Exception(f'Synchronization error, please rerun simulation')

    ret_pipe.send(vcd_conv.json_vcd)


def dtype_tree(dtype):
    from pygears.typing import Queue, Array, Tuple
    if issubclass(dtype, Array):
        return {'name': repr(dtype), 'subtypes': [dtype_tree(t) for t in dtype]}
    elif issubclass(dtype, (Queue, Array, Tuple)):
        return {
            'name': repr(dtype),
            'subtypes': [dtype_tree(t) for t in dtype],
            'keys': dtype.fields
        }
    else:
        return {'name': repr(dtype)}


def dump_json_graph(top):

    nodes = {}
    ports = {}
    connections = {}

    class NodeYielder(HierYielderBase):
        def Gear(self, node):
            yield node
            return not node.hierarchical

    for node in NodeYielder().visit(find(top)):
        in_ports = []
        for p in node.in_ports:
            ports[p.name] = {
                'basename': p.basename,
                'name': p.name,
                'dtype': repr(p.dtype),
                'index': p.index
            }
            in_ports.append(p.name)

        out_ports = []
        for p in node.out_ports:
            ports[p.name] = {
                'basename': p.basename,
                'name': p.name,
                'dtype': repr(p.dtype),
                'index': p.index
            }
            out_ports.append(p.name)

        node_json = {
            'basename': node.basename,
            'name': node.name,
            'definition': 'None' if node.definition is None else
            f'{node.definition.func.__module__}.{node.definition.func.__name__}',
            'in_ports': in_ports,
            'out_ports': out_ports
        }

        nodes[node.name] = node_json

    for node in NodeYielder().visit(find(top)):
        nodes[node.name]['child'] = []
        if node.hierarchical:
            for c in node.child:
                nodes[node.name]['child'].append(c.name)

    for node in NodeYielder().visit(find(top)):
        for p in node.in_ports + node.out_ports:
            if p.consumer is None:
                continue

            i = p.consumer
            connections[i.name] = {
                'producer': i.producer.name,
                'dtype': repr(p.dtype),
                'dtype_tree': dtype_tree(p.dtype),
                'name': i.name,
                'consumers': [pc.name for pc in i.consumers if not isinstance(pc, HDLConsumer)]
            }

            # if not all(isinstance(p, HDLConsumer) for p in i.consumers):
            ports[i.producer.name]['consumer'] = i.name
            for p in i.consumers:
                if isinstance(p, HDLConsumer):
                    continue

                ports[p.name]['producer'] = i.name

    return {'gears': nodes, 'ports': ports, 'connections': connections}


class WebSim(SimExtend):
    @inject
    def __init__(self, trace_fn='pygears.json', outdir=Inject('results-dir')):
        super().__init__()
        self.outdir = outdir
        self.trace_fn = os.path.abspath(os.path.join(self.outdir, trace_fn))
        # atexit.register(self.finish)

        self.vcd_fn = os.path.abspath(os.path.join(self.outdir, 'pygears.vcd'))
        self.finished = False

        self.json_vcd = {}
        manager = multiprocessing.Manager()
        self.json_vcd = manager.dict()

    def register_vcd_worker(self, vcd_fn, top):
        qin, qout = multiprocessing.Pipe(duplex=False)
        self.qin.append(qin)

        p = multiprocessing.Process(target=vcd_to_json, args=(vcd_fn, self.finish_event, qout, top))
        self.p.append(p)

    def before_run(self, sim):
        self.finish_event = multiprocessing.Event()

        self.qin = []
        self.p = []
        for m in find_cosim_modules():
            self.register_vcd_worker(m.trace_fn, top=m.gear.parent)

        self.register_vcd_worker(self.vcd_fn, find('/'))

        for p in self.p:
            p.start()

    def sim_vcd_to_json(self):
        graph = dump_json_graph('/')

        self.finish_event.set()
        json_vcds = [qin.recv() for qin in self.qin]

        for p in self.p:
            p.join()

        changes = []
        for json_vcd in json_vcds:
            for p_name in json_vcd:
                p = find(p_name)
                if isinstance(p, Intf):
                    changes.append({'channelName': p_name, 'changes': json_vcd[p_name]})
                else:
                    changes.append({'channelName': p.producer.name, 'changes': json_vcd[p_name]})

        return {
            'graphInfo': graph,
            'simulationChanges': {
                'startCycle': 0,
                'endCycle': timestep(),
                'channelChanges': changes
            }
        }

    def finish(self):
        if not self.finished:
            json_out = self.sim_vcd_to_json()
            import json
            json.dump(json_out, open(self.trace_fn, 'w'), separators=(',', ':'))
            # json.dump(json_out, open(self.trace_fn, 'w'))
            # json.dump(json_out, open(self.trace_fn, 'w'), indent=4)

            self.finished = True

    def after_cleanup(self, sim):
        self.finish()


def websim_activate(var, val):
    if val:
        if WebSim not in reg['sim/extens']:
            reg['debug/expand_trace_data'] = False
            reg['debug/trace_end_cycle_dump'] = False
            reg['sim/extens'].append(WebSim)
    else:
        if WebSim in reg['sim/extens']:
            reg['debug/expand_trace_data'] = True
            reg['debug/trace_end_cycle_dump'] = True
            del reg['sim/extens'][reg['sim/extens'].index(WebSim)]


class WebSimPlugin(HDLPlugin):
    @classmethod
    def bind(cls):
        reg.confdef('debug/webviewer', setter=websim_activate, default=False)



import pkg_resources
import pkgutil
import importlib

import sys
if sys.version_info[0] < 3 or sys.version_info[1] < 6:
    raise Exception("Must be using Python 3.6 and above")

__version__ = pkg_resources.get_distribution("pygears").version

import sys
from asyncio import CancelledError as GearDone
from asyncio.queues import QueueEmpty as IntfEmpty

import pygears.typing

# form pygears.conf import only for legacy compatibility
from pygears.conf import PluginBase, clear, reset, MultiAlternativeError
from pygears.conf import pygears_excepthook, TraceLevel, reg
import pygears.conf
import pygears.entry

from pygears.util.find import find

from pygears.typing import TypeMatchError
from pygears.core.gear import module
from pygears.core.gear_decorator import gear, alternative
import pygears.core.gear_inst

from pygears.core.intf import Intf

from pygears.core.datagear import datagear

import pygears.sim
from pygears.sim import sim

import pygears.lib
import pygears.typing.pprint

# import os
# from pygears.registry import load_plugin_folder
# load_plugin_folder(os.path.join(os.path.dirname(__file__), 'lib'))

import pygears.hdl

non_plugin_pkgs = ['pygears_live', 'pygears_tools']
plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg in pkgutil.iter_modules()
    if name.startswith('pygears_') and name not in non_plugin_pkgs
}

from pygears.conf.custom_settings import load_rc, print_registry
load_rc('.pygears')
# print_registry()

sys.excepthook = pygears_excepthook

__all__ = [
    'TraceLevel', 'reg', 'gear', 'alternative', 'clear', 'Intf', 'PluginBase', 'find',
    'MultiAlternativeError', 'GearDone', 'IntfEmpty', 'module', 'TypeMatchError',
    'datagear'
]



from pygears.core.gear import Gear
from pygears import reg
from pygears.conf.trace import TraceLevel
from .ast import FuncContext, GearContext, form_hls_syntax_error, visit_ast
from .ast.utils import get_function_ast
from .ast.inline import removed_unused_vars
from .passes import (remove_dead_code, infer_registers, schedule, handle_generators, resolve_gear_calls,
                     find_called_funcs, loop_unfold, detect_loops)

from .passes.schedule import RebuildStateIR

from .passes.inline_cfg import VarScope

from .debug import hls_debug
from .debug import print_gear_parse_intro
from . import cfg as cfgutil


def translate_gear(gear: Gear):
    # hls_enable_debug_log()
    hls_debug(title=f'Translating: {gear.name}')

    exec_context = reg['gear/exec_context']

    parent = reg['gear/current_module']
    # For the code that inspects gear via module() call
    reg['gear/current_module'] = gear

    body_ast = get_function_ast(gear.func)

    print_gear_parse_intro(gear, body_ast)
    ctx = GearContext(gear)

    res = process(body_ast, ctx)

    if reg['trace/level'] == TraceLevel.user:
        exc_value = None
        try:
            res = transform(res, ctx)
        except Exception as e:
            exc_value, tb = form_hls_syntax_error(ctx, e)

        if exc_value is not None:
            raise exc_value.with_traceback(tb)
    else:
        res = transform(res, ctx)

    reg['gear/exec_context'] = exec_context
    reg['gear/current_module'] = parent
    return ctx, res


def process(body_ast, ctx):
    # hls_enable_debug_log()
    reg['gear/exec_context'] = 'hls'
    reg['hls/ctx'] = [ctx]
    return visit_ast(body_ast, ctx)


def transform(modblock, ctx: GearContext):
    hls_debug(modblock, 'Initial')

    modblock, cfg, reaching = cfgutil.forward(modblock, cfgutil.ReachingDefinitions())
    ctx.reaching = {id(n.value): v for n, v in reaching.items()}

    detect_loops(modblock, ctx)

    modblock = infer_registers(modblock, ctx)

    # modblock = resolve_gear_calls(modblock, ctx)
    # hls_debug(modblock, 'Resolve Gear Calls')

    modblock = handle_generators(modblock, ctx)
    hls_debug(modblock, 'Handle Generators')

    # print(modblock)

    modblock = loop_unfold(modblock, ctx)
    hls_debug(modblock, 'Handle Generators')
    # print(modblock)

    modblock, cfg, reaching = cfgutil.forward(modblock, cfgutil.ReachingDefinitions())
    ctx.reaching = {id(n.value): v for n, v in reaching.items()}

    modblock = schedule(cfg, ctx)

    modblock = remove_dead_code(modblock, ctx)
    hls_debug(modblock, 'Remove Dead Code')

    modblock = removed_unused_vars(modblock, ctx)

    compile_funcs(modblock, ctx)

    return modblock


def compile_funcs(modblock, ctx):
    called_funcs = find_called_funcs(modblock, ctx)

    added = bool(called_funcs)

    active_funcs = set()
    while added:
        added = False

        functions = ctx.functions.copy()

        for f_ast, f_ctx in functions.values():
            if (f_ast.name in active_funcs) or (f_ast.name not in called_funcs):
                continue

            active_funcs.add(f_ctx.funcref.name)
            funcblock = transform_func(f_ast, f_ctx)

            func_called_funcs = find_called_funcs(funcblock, f_ctx)
            if func_called_funcs:
                added = True
                called_funcs.update(func_called_funcs)

            modblock.funcs.append((funcblock, f_ctx))


def transform_func(funcblock, ctx: FuncContext):
    funcblock = resolve_gear_calls(funcblock, ctx)
    # hls_debug(funcblock, 'Resolve Gear Calls')

    funcblock = handle_generators(funcblock, ctx)
    hls_debug(funcblock, 'Handle Generators')

    cfg = cfgutil.CFG.build_cfg(funcblock)
    VarScope(ctx).visit(cfg.entry)

    v = RebuildStateIR()
    v.visit(cfg.entry)
    funcblock = cfg.entry.value

    # funcblock = inline(funcblock, ctx)

    funcblock = remove_dead_code(funcblock, ctx)

    funcblock = removed_unused_vars(funcblock, ctx)

    return funcblock



from .. import ir
from .. import ir_utils
from ..debug import hls_debug

from .dead_code import remove_dead_code, find_called_funcs
from .register import infer_registers
from .schedule import schedule
from .generators import handle_generators
from .gears import resolve_gear_calls
from .unfolder import loop_unfold, detect_loops

__all__ = [
    'infer_exit_cond', 'remove_dead_code',
    'infer_registers', 'infer_in_cond', 'resolve_gear_calls'
]
